from redis_om import JsonModel
from redis_om.model.model import Field


class EnvAgentComboStorage(JsonModel):
    env_id: str = Field(default_factory=lambda: "", index=True)
    agent_ids: list[str] = Field(default_factory=lambda: [], index=True)



import csv
import json
from typing import Any

from pydantic import BaseModel, Field

from .env_agent_combo_storage import EnvAgentComboStorage
from .logs import EpisodeLog
from .persistent_profile import (
    AgentProfile,
    EnvironmentProfile,
    RelationshipProfile,
)


class TwoAgentEpisodeWithScenarioBackgroundGoals(BaseModel):
    episode_id: str = Field(required=True)
    environment_id: str = Field(required=True)
    agent_ids: list[str] = Field(required=True)
    experiment_tag: str = Field(required=True)
    experiment_model_name_pairs: list[str] = Field(required=True)
    raw_messages: list[list[tuple[str, str, str]]] = Field(required=True)
    raw_rewards: list[tuple[float, dict[str, float]] | float] = Field(required=True)
    raw_rewards_prompt: str = Field(required=True)
    scenario: str = Field(required=True)
    codename: str = Field(required=True)
    agents_background: dict[str, str] = Field(required=True)
    social_goals: dict[str, str] = Field(required=True)
    social_interactions: str = Field(required=True)
    reasoning: str = Field(required=False)
    rewards: list[tuple[float, dict[str, float]]] = Field(required=False)


class AgentProfileWithPersonalInformation(BaseModel):
    agent_id: str = Field(required=True)
    first_name: str = Field(required=True)
    last_name: str = Field(required=True)
    age: int = Field(required=True)
    occupation: str = Field(required=True)
    gender: str = Field(required=True)
    gender_pronoun: str = Field(required=True)
    public_info: str = Field(required=True)
    big_five: str = Field(required=True)
    moral_values: list[str] = Field(required=True)
    schwartz_personal_values: list[str] = Field(required=True)
    personality_and_values: str = Field(required=True)
    decision_making_style: str = Field(required=True)
    secret: str = Field(required=True)
    mbti: str = Field(required=True)
    model_id: str = Field(required=True)


class EnvironmentProfileWithTwoAgentRequirements(BaseModel):
    env_id: str = Field(required=True)
    codename: str = Field(required=True)
    source: str = Field(required=True)
    scenario: str = Field(required=True)
    agent_goals: list[str] = Field(required=True)
    relationship: str = Field(required=True)
    age_constraint: str = Field(required=True)
    occupation_constraint: str = Field(required=True)
    agent_constraint: str | None = Field(required=False)


class RelationshipProfileBetweenTwoAgents(BaseModel):
    relationship_id: str = Field(required=True)
    agent1_id: str = Field(required=True)
    agent2_id: str = Field(required=True)
    relationship: str = Field(required=True)
    background_story: str = Field(required=True)


class EnvAgentComboStorageWithID(BaseModel):
    combo_id: str = Field(default_factory=lambda: "", index=True)
    env_id: str = Field(default_factory=lambda: "", index=True)
    agent_ids: list[str] = Field(default_factory=lambda: [], index=True)


def _map_gender_to_adj(gender: str) -> str:
    gender_to_adj = {
        "Man": "male",
        "Woman": "female",
        "Nonbinary": "nonbinary",
    }
    if gender:
        return gender_to_adj[gender]
    else:
        return ""


def get_rewards_from_episode(
    episode: EpisodeLog,
) -> list[tuple[float, dict[str, float]]]:
    assert (
        len(episode.rewards) == 2
        and (not isinstance(episode.rewards[0], float))
        and (not isinstance(episode.rewards[1], float))
    )
    return [episode.rewards[0], episode.rewards[1]]


def get_scenario_from_episode(episode: EpisodeLog) -> str:
    """Get the scenario from the episode.

    Args:
        episode (EpisodeLog): The episode.

    Returns:
        str: The scenario.
    """
    return EnvironmentProfile.get(pk=episode.environment).scenario


def get_codename_from_episode(episode: EpisodeLog) -> str:
    """Get the codename from the episode.

    Args:
        episode (EpisodeLog): The episode.

    Returns:
        str: The codename.
    """
    return EnvironmentProfile.get(pk=episode.environment).codename


def get_agents_background_from_episode(episode: EpisodeLog) -> dict[str, str]:
    """Get the agents' background from the episode.

    Args:
        episode (EpisodeLog): The episode.

    Returns:
        list[str]: The agents' background.
    """
    agents = [AgentProfile.get(pk=agent) for agent in episode.agents]

    return {
        f"{profile.first_name} {profile.last_name}": f"{profile.first_name} {profile.last_name} is a {profile.age}-year-old {_map_gender_to_adj(profile.gender)} {profile.occupation.lower()}. {profile.gender_pronoun} pronouns. {profile.public_info} Personality and values description: {profile.personality_and_values} {profile.first_name}'s secrets: {profile.secret}"
        for profile in agents
    }


def get_agent_name_to_social_goal_from_episode(
    episode: EpisodeLog,
) -> dict[str, str]:
    agents = [AgentProfile.get(agent) for agent in episode.agents]
    agent_names = [agent.first_name + " " + agent.last_name for agent in agents]
    environment = EnvironmentProfile.get(episode.environment)
    agent_goals = {
        agent_names[0]: environment.agent_goals[0],
        agent_names[1]: environment.agent_goals[1],
    }
    return agent_goals


def get_social_interactions_from_episode(
    episode: EpisodeLog,
) -> str:
    assert isinstance(episode.tag, str)
    list_of_social_interactions = episode.render_for_humans()[1]
    if len(list_of_social_interactions) < 3:
        return ""
    if "script" in episode.tag.split("_"):
        overall_social_interaction = list_of_social_interactions[1:-3]
    else:
        overall_social_interaction = list_of_social_interactions[0:-3]
        # only get the sentence after "Conversation Starts:\n\n"
        starter_msg_list = overall_social_interaction[0].split(
            "Conversation Starts:\n\n"
        )
        if len(starter_msg_list) < 3:
            overall_social_interaction = list_of_social_interactions[1:-3]
            # raise ValueError("The starter message is not in the expected format")
        else:
            overall_social_interaction[0] = starter_msg_list[-1]
    return "\n\n".join(overall_social_interaction)


def _serialize_data_to_csv(data: dict[str, list[Any]], csv_file_path: str) -> None:
    """
    Serialize data to a csv file without pandas
    """
    max_length = max(len(lst) for lst in data.values())

    # Create and write to the csv file
    with open(csv_file_path, "w", newline="") as file:
        writer = csv.writer(file)
        # Write the header
        writer.writerow(data.keys())

        # Write the data rows
        for i in range(max_length):
            # Create a row for each index in the longest list
            row = [data[key][i] if i < len(data[key]) else "" for key in data.keys()]
            writer.writerow(row)


def episodes_to_csv(
    episodes: list[EpisodeLog], csv_file_path: str = "episodes.csv"
) -> None:
    """Save episodes to a csv file.

    Args:
        episodes (list[EpisodeLog]): List of episodes.
        filepath (str, optional): The file path. Defaults to "episodes.csv".
    """
    data: dict[str, list[Any]] = {
        "episode_id": [episode.pk for episode in episodes],
        "environment_id": [episode.environment for episode in episodes],
        "agent_ids": [episode.agents for episode in episodes],
        "experiment_tag": [episode.tag for episode in episodes],
        "experiment_model_name_pairs": [episode.models for episode in episodes],
        "raw_messages": [episode.messages for episode in episodes],
        "raw_rewards_prompt": [episode.rewards_prompt for episode in episodes],
        "raw_rewards": [episode.rewards for episode in episodes],
        "scenario": [get_scenario_from_episode(episode) for episode in episodes],
        "codename": [get_codename_from_episode(episode) for episode in episodes],
        "agents_background": [
            get_agents_background_from_episode(episode) for episode in episodes
        ],
        "social_goals": [
            get_agent_name_to_social_goal_from_episode(episode) for episode in episodes
        ],
        "social_interactions": [
            get_social_interactions_from_episode(episode) for episode in episodes
        ],
        "reasoning": [episode.reasoning for episode in episodes],
        "rewards": [get_rewards_from_episode(episode) for episode in episodes],
    }

    # Serialize data to a csv file without pandas
    _serialize_data_to_csv(data, csv_file_path)


def episodes_to_jsonl(
    episodes: list[EpisodeLog], jsonl_file_path: str = "episodes.jsonl"
) -> None:
    """Save episodes to a json file.

    Args:
        episodes (list[EpisodeLog]): List of episodes.
        filepath (str, optional): The file path. Defaults to "episodes.json".
    """
    with open(jsonl_file_path, "w") as f:
        for episode in episodes:
            data = TwoAgentEpisodeWithScenarioBackgroundGoals(
                episode_id=episode.pk,
                environment_id=episode.environment,
                agent_ids=episode.agents,
                experiment_tag=episode.tag,
                experiment_model_name_pairs=episode.models,
                raw_messages=episode.messages,
                raw_rewards_prompt=episode.rewards_prompt,
                raw_rewards=episode.rewards,
                scenario=get_scenario_from_episode(episode),
                codename=get_codename_from_episode(episode),
                agents_background=get_agents_background_from_episode(episode),
                social_goals=get_agent_name_to_social_goal_from_episode(episode),
                social_interactions=get_social_interactions_from_episode(episode),
                reasoning=episode.reasoning,
                rewards=get_rewards_from_episode(episode),
            )
            json.dump(dict(data), f)
            f.write("\n")


def agentprofiles_to_csv(
    agent_profiles: list[AgentProfile],
    csv_file_path: str = "agent_profiles.csv",
) -> None:
    """Save agent profiles to a csv file.

    Args:
        agent_profiles (list[AgentProfile]): List of agent profiles.
        filepath (str, optional): The file path. Defaults to "agent_profiles.csv".
    """
    data: dict[str, list[Any]] = {
        "agent_id": [profile.pk for profile in agent_profiles],
        "first_name": [profile.first_name for profile in agent_profiles],
        "last_name": [profile.last_name for profile in agent_profiles],
        "age": [profile.age for profile in agent_profiles],
        "occupation": [profile.occupation for profile in agent_profiles],
    }

    # Serialize data to a csv file without pandas
    _serialize_data_to_csv(data, csv_file_path)


def agentprofiles_to_jsonl(
    agent_profiles: list[AgentProfile],
    jsonl_file_path: str = "agent_profiles.jsonl",
) -> None:
    """Save agent profiles to a json file.

    Args:
        agent_profiles (list[AgentProfile]): List of agent profiles.
        filepath (str, optional): The file path. Defaults to "agent_profiles.json".
    """
    with open(jsonl_file_path, "w") as f:
        for profile in agent_profiles:
            data = AgentProfileWithPersonalInformation(
                agent_id=profile.pk,
                first_name=profile.first_name,
                last_name=profile.last_name,
                age=profile.age,
                occupation=profile.occupation,
                gender=profile.gender,
                gender_pronoun=profile.gender_pronoun,
                public_info=profile.public_info,
                big_five=profile.big_five,
                moral_values=profile.moral_values,
                schwartz_personal_values=profile.schwartz_personal_values,
                personality_and_values=profile.personality_and_values,
                decision_making_style=profile.decision_making_style,
                secret=profile.secret,
                mbti=profile.mbti,
                model_id=profile.model_id,
            )
            json.dump(dict(data), f)
            f.write("\n")


def environmentprofiles_to_csv(
    environment_profiles: list[EnvironmentProfile],
    csv_file_path: str = "environment_profiles.csv",
) -> None:
    """Save environment profiles to a csv file.

    Args:
        environment_profiles (list[EnvironmentProfile]): List of environment profiles.
        filepath (str, optional): The file path. Defaults to "environment_profiles.csv".
    """
    data: dict[str, list[Any]] = {
        "env_id": [profile.pk for profile in environment_profiles],
        "codename": [profile.codename for profile in environment_profiles],
        "source": [profile.source for profile in environment_profiles],
        "scenario": [profile.scenario for profile in environment_profiles],
        "agent_goals": [profile.agent_goals for profile in environment_profiles],
        "relationship": [profile.relationship for profile in environment_profiles],
        "age_constraint": [profile.age_constraint for profile in environment_profiles],
        "occupation_constraint": [
            profile.occupation_constraint for profile in environment_profiles
        ],
        "agent_constraint": [
            profile.agent_constraint for profile in environment_profiles
        ],
    }

    # Serialize data to a csv file without pandas
    _serialize_data_to_csv(data, csv_file_path)


def environmentprofiles_to_jsonl(
    environment_profiles: list[EnvironmentProfile],
    jsonl_file_path: str = "environment_profiles.jsonl",
) -> None:
    """Save environment profiles to a json file.

    Args:
        environment_profiles (list[EnvironmentProfile]): List of environment profiles.
        filepath (str, optional): The file path. Defaults to "environment_profiles.json".
    """
    with open(jsonl_file_path, "w") as f:
        for profile in environment_profiles:
            data = EnvironmentProfileWithTwoAgentRequirements(
                env_id=profile.pk,
                codename=profile.codename,
                source=profile.source,
                scenario=profile.scenario,
                agent_goals=profile.agent_goals,
                relationship=profile.relationship,
                age_constraint=profile.age_constraint,
                occupation_constraint=profile.occupation_constraint,
                agent_constraint=profile.agent_constraint
                if profile.agent_constraint
                else "none",
            )
            json.dump(dict(data), f)
            f.write("\n")


def relationshipprofiles_to_csv(
    relationship_profiles: list[RelationshipProfile],
    csv_file_path: str = "relationship_profiles.csv",
) -> None:
    """Save relationship profiles to a csv file.

    Args:
        relationship_profiles (list[RelationshipProfile]): List of relationship profiles.
        filepath (str, optional): The file path. Defaults to "relationship_profiles.csv".
    """
    data: dict[str, list[Any]] = {
        "relationship_id": [profile.pk for profile in relationship_profiles],
        "agent1_id": [profile.agent_1_id for profile in relationship_profiles],
        "agent2_id": [profile.agent_2_id for profile in relationship_profiles],
        "relationship": [profile.relationship for profile in relationship_profiles],
        "background_story": [
            profile.background_story for profile in relationship_profiles
        ],
    }

    # Serialize data to a csv file without pandas
    _serialize_data_to_csv(data, csv_file_path)


def envagnetcombostorage_to_csv(
    env_agent_combo_storages: list[EnvAgentComboStorage],
    csv_file_path: str = "env_agent_combo_storages.csv",
) -> None:
    """Save environment-agent combo storages to a csv file.

    Args:
        env_agent_combo_storages (list[EnvAgentComboStorage]): List of environment-agent combo storages.
        filepath (str, optional): The file path. Defaults to "env_agent_combo_storages.csv".
    """
    data: dict[str, list[Any]] = {
        "combo_id": [storage.pk for storage in env_agent_combo_storages],
        "env_id": [storage.env_id for storage in env_agent_combo_storages],
        "agent_ids": [storage.agent_ids for storage in env_agent_combo_storages],
    }

    # Serialize data to a csv file without pandas
    _serialize_data_to_csv(data, csv_file_path)


def relationshipprofiles_to_jsonl(
    relationship_profiles: list[RelationshipProfile],
    jsonl_file_path: str = "relationship_profiles.jsonl",
) -> None:
    """Save relationship profiles to a json file.

    Args:
        relationship_profiles (list[RelationshipProfile]): List of relationship profiles.
        filepath (str, optional): The file path. Defaults to "relationship_profiles.json".
    """
    with open(jsonl_file_path, "w") as f:
        for profile in relationship_profiles:
            data = RelationshipProfileBetweenTwoAgents(
                relationship_id=profile.pk,
                agent1_id=profile.agent_1_id,
                agent2_id=profile.agent_2_id,
                relationship=profile.relationship,
                background_story=profile.background_story,
            )
            json.dump(dict(data), f)
            f.write("\n")


def envagnetcombostorage_to_jsonl(
    env_agent_combo_storages: list[EnvAgentComboStorage],
    jsonl_file_path: str = "env_agent_combo_storages.jsonl",
) -> None:
    """Save environment-agent combo storages to a json file.

    Args:
        env_agent_combo_storages (list[EnvAgentComboStorage]): List of environment-agent combo storages.
        filepath (str, optional): The file path. Defaults to "env_agent_combo_storages.json".
    """
    with open(jsonl_file_path, "w") as f:
        for storage in env_agent_combo_storages:
            data = EnvAgentComboStorageWithID(
                combo_id=storage.pk,
                env_id=storage.env_id,
                agent_ids=storage.agent_ids,
            )
            json.dump(dict(data), f)
            f.write("\n")


def jsonl_to_episodes(
    jsonl_file_path: str,
) -> list[EpisodeLog]:
    """Load episodes from a jsonl file.

    Args:
        jsonl_file_path (str): The file path.

    Returns:
        list[EpisodeLog]: List of episodes.
    """
    episodes = []
    with open(jsonl_file_path, "r") as f:
        for line in f:
            data = json.loads(line)
            episode = EpisodeLog(
                pk=data["episode_id"],
                environment=data["environment_id"],
                agents=data["agent_ids"],
                tag=data["experiment_tag"],
                models=data["experiment_model_name_pairs"],
                messages=data["raw_messages"],
                reasoning=data["reasoning"],
                rewards=data["raw_rewards"],
                rewards_prompt=data["raw_rewards_prompt"],
            )
            episodes.append(episode)
    return episodes


def jsonl_to_agentprofiles(
    jsonl_file_path: str,
) -> list[AgentProfile]:
    """Load agent profiles from a jsonl file.

    Args:
        jsonl_file_path (str): The file path.

    Returns:
        list[AgentProfile]: List of agent profiles.
    """
    agent_profiles = []
    with open(jsonl_file_path, "r") as f:
        for line in f:
            data = json.loads(line)
            agent_profile = AgentProfile(
                pk=data["agent_id"],
                first_name=data["first_name"],
                last_name=data["last_name"],
                age=data["age"],
                occupation=data["occupation"],
                gender=data["gender"],
                gender_pronoun=data["gender_pronoun"],
                public_info=data["public_info"],
                big_five=data["big_five"],
                moral_values=data["moral_values"],
                schwartz_personal_values=data["schwartz_personal_values"],
                personality_and_values=data["personality_and_values"],
                decision_making_style=data["decision_making_style"],
                secret=data["secret"],
                model_id=data["model_id"],
                mbti=data["mbti"],
            )
            agent_profiles.append(agent_profile)
    return agent_profiles


def jsonl_to_environmentprofiles(
    jsonl_file_path: str,
) -> list[EnvironmentProfile]:
    """Load environment profiles from a jsonl file.

    Args:
        jsonl_file_path (str): The file path.

    Returns:
        list[EnvironmentProfile]: List of environment profiles.
    """
    environment_profiles = []
    with open(jsonl_file_path, "r") as f:
        for line in f:
            data = json.loads(line)
            environment_profile = EnvironmentProfile(
                pk=data["env_id"],
                codename=data["codename"],
                source=data["source"],
                scenario=data["scenario"],
                agent_goals=data["agent_goals"],
                relationship=data["relationship"],
                age_constraint=data["age_constraint"],
                occupation_constraint=data["occupation_constraint"],
                agent_constraint=data["agent_constraint"]
                if data["agent_constraint"] != "none"
                else None,
            )
            environment_profiles.append(environment_profile)
    return environment_profiles


def jsonl_to_relationshipprofiles(
    jsonl_file_path: str,
) -> list[RelationshipProfile]:
    """Load relationship profiles from a jsonl file.

    Args:
        jsonl_file_path (str): The file path.

    Returns:
        list[RelationshipProfileBetweenTwoAgents]: List of relationship profiles.
    """
    relationship_profiles = []
    with open(jsonl_file_path, "r") as f:
        for line in f:
            data = json.loads(line)
            relationship_profile = RelationshipProfile(
                pk=data["relationship_id"],
                agent_1_id=data["agent1_id"],
                agent_2_id=data["agent2_id"],
                relationship=data["relationship"],
                background_story=data["background_story"],
            )
            relationship_profiles.append(relationship_profile)
    return relationship_profiles


def jsonl_to_envagnetcombostorage(
    jsonl_file_path: str,
) -> list[EnvAgentComboStorage]:
    """Load environment-agent combo storages from a jsonl file.

    Args:
        jsonl_file_path (str): The file path.

    Returns:
        list[EnvAgentComboStorageWithID]: List of environment-agent combo storages.
    """
    env_agent_combo_storages = []
    with open(jsonl_file_path, "r") as f:
        for line in f:
            data = json.loads(line)
            env_agent_combo_storage = EnvAgentComboStorage(
                pk=data["combo_id"],
                env_id=data["env_id"],
                agent_ids=data["agent_ids"],
            )
            env_agent_combo_storages.append(env_agent_combo_storage)
    return env_agent_combo_storages



from .generate import (
    EnvResponse,
    LLM_Name,
    agenerate_env_profile,
    agenerate,
    agenerate_action,
)

from .sync import (
    generate,
    generate_action,
)

__all__ = [
    "EnvResponse",
    "agenerate_env_profile",
    "LLM_Name",
    "agenerate",
    "agenerate_action",
    "generate",
    "generate_action",
]



from datetime import datetime
import requests
import rich
from sotopia.database.persistent_profile import EnvironmentList
import asyncio
import logging
import json
import math
import numpy as np
from itertools import chain
from collections import defaultdict
from typing import cast, OrderedDict

from logging import FileHandler
from rich.logging import RichHandler

from tqdm import tqdm

from sotopia.agents import LLMAgent
from sotopia.database import (
    AgentProfile,
    EnvAgentComboStorage,
    EnvironmentProfile,
    EpisodeLog,
)
from sotopia.database.serialization import get_rewards_from_episode
from sotopia.envs.evaluators import (
    EvaluationForTwoAgents,
    ReachGoalLLMEvaluator,
    RuleBasedTerminatedEvaluator,
    SotopiaDimensions,
)
from sotopia.envs.parallel import ParallelSotopiaEnv
from sotopia.generation_utils.generate import LLM_Name
from sotopia.messages import AgentAction, Observation
from sotopia.samplers import (
    BaseSampler,
    EnvAgentCombo,
)
from sotopia.server import run_async_server

import typer
from pathlib import Path
from ..app import app
import os

# date and message only
FORMAT = "%(asctime)s - %(levelname)s - %(name)s - %(message)s"

logging.basicConfig(
    level=20,
    format=FORMAT,
    datefmt="[%X]",
    handlers=[
        RichHandler(),
    ],
)

default_model_list: list[str] = [
    "gpt-4o",
]
dimension_range_mapping = OrderedDict(
    {
        "social_rules": ["SOC", [-10, 0]],
        "secret": ["SEC", [-10, 0]],
        "financial_and_material_benefits": ["FIN", [-5, 5]],
        "relationship": ["REL", [-5, 5]],
        "knowledge": ["KNO", [0, 10]],
        "goal": ["GOAL", [0, 10]],
        "believability": ["BEL", [0, 10]],
    }
)


def get_avg_reward(
    episodes: list[EpisodeLog], model_name: str
) -> dict[str, tuple[float, float]]:
    """
    input: list of EpisodeLog, model_name

    return: dictionary of {dimension: (avg_reward, margin_of_error (in 95% confidence interval))}, plus the distinct setting number and episode count (in the same format, but with 0 margin of error)
    """
    rewards_dict = defaultdict(
        list
    )  # {pk: [rewards]}, {pk}_{i} denotes the i-th agent is the test agent
    avg_reward_dict = {}
    avg_margin_dict = {}
    avg_variance_dict = {}
    for episode in episodes:
        assert episode.models is not None, "episode.models should not be None"
        if episode.models[1] == model_name:
            reward = get_rewards_from_episode(episode)[0][1]
            rewards_dict[f"{episode.environment}_0"].append(reward)
        else:
            reward = get_rewards_from_episode(episode)[1][1]
            rewards_dict[f"{episode.environment}_1"].append(reward)
    dimensions = list(rewards_dict.values())[0][0].keys()

    def calc_variance(local_rewards_list: list[dict[str, float]]) -> dict[str, float]:
        # get the variance within a list, discarded
        local_var_reward_dict = {}
        local_dimensions = local_rewards_list[0].keys()
        assert set(local_dimensions) == set(dimensions), "dimensions should be the same"
        for dimension in local_dimensions:
            rewards = [reward[dimension] for reward in local_rewards_list]
            avg_reward = sum(rewards) / len(rewards)
            if len(rewards) == 1:
                variance = 0.0
            else:
                variance = sum([(reward - avg_reward) ** 2 for reward in rewards]) / (
                    len(rewards) - 1
                )
            local_var_reward_dict[dimension] = variance

        return local_var_reward_dict

    def calc_average(list_to_average: list[float]) -> float:
        return sum(list_to_average) / len(list_to_average)

    rewards_list = list(chain(*rewards_dict.values()))

    variance_reward_list = [calc_variance(rewards) for rewards in rewards_dict.values()]
    for dimension in rewards_list[0].keys():
        avg_reward_dict[dimension] = calc_average(
            [reward[dimension] for reward in rewards_list]
        )
        avg_variance_dict[dimension] = calc_average(
            [variance[dimension] for variance in variance_reward_list]
        )  # average the variances for an estimation of the variance

    for dimension in rewards_list[0].keys():
        # calculate the margin of error by the averaged mean and variance
        avg_variance = avg_variance_dict[dimension]

        combined_variance = avg_variance
        combined_sem = math.sqrt(
            combined_variance / len(rewards_list)
        )  # sem = sqrt(variance / n), we use the averaged variance under different settings

        confidence_level = 0.95
        t_samples = np.random.standard_t(df=len(rewards_list) - 1, size=1000000)

        overall_t_value = np.percentile(
            t_samples, 100 * (1 - (1 - confidence_level) / 2)
        )

        margin = overall_t_value * combined_sem
        avg_margin_dict[dimension] = margin

    return_rewards_dict = {
        key: (avg_reward_dict[key], avg_margin_dict[key])
        for key in avg_reward_dict.keys()
    }
    return_rewards_dict = {
        **return_rewards_dict,
        "setting_num": (float(len(variance_reward_list)), 0.0),
        "episode_count": (float(len(rewards_list)), 0.0),
    }

    return return_rewards_dict


def initialize_benchmark_combo(url: str) -> list[EnvAgentComboStorage]:
    if url:
        response = requests.get(url)
        # Check if the request was successful
        if response.status_code == 200:
            data = response.json()
            print("Data fetched successfully")
        else:
            print(f"Failed to fetch data. Status code: {response.status_code}")
            raise ValueError("Failed to fetch data")
        list_of_env_agent_combo_storage = []
        for combo in data:
            env_agent_combo_storage = EnvAgentComboStorage(
                env_id=combo["env_id"], agent_ids=combo["agent_ids"]
            )
            list_of_env_agent_combo_storage.append(env_agent_combo_storage)
    else:
        list_of_env_agent_combo_storage = EnvAgentComboStorage.find().all()  # type: ignore
    return list_of_env_agent_combo_storage


def _set_up_logs(
    *,
    log_file_level: int = logging.DEBUG,
    log_rich_level: int = logging.INFO,
    log_format: str = "%(asctime)s - %(levelname)s - %(name)s - %(message)s",
    log_file: str = datetime.now().strftime("./logs/%H_%M_%d_%m_%Y.log"),
    print_logs: bool = False,
) -> None:
    # date and message only
    logging_path = Path(log_file)
    logging_path.parent.mkdir(parents=True, exist_ok=True)

    logging.basicConfig(
        level=log_file_level,
        format=log_format,
        datefmt="[%X]",
        handlers=[
            FileHandler(logging_path),
            RichHandler(level=log_rich_level) if print_logs else RichHandler(level=100),
        ],
    )


def preprocess_episode_data(
    episode_list_with_tag: list[EpisodeLog],
) -> dict[tuple[str, tuple[str, ...], tuple[str, ...]], bool]:
    """Preprocess episodes into a dictionary for quick lookup."""
    episode_dict = {}
    for episode in episode_list_with_tag:
        assert isinstance(episode, EpisodeLog), "episode should be an EpisodeLog"
        assert episode.models is not None, "episode.models should not be None"
        episode_key = (
            episode.environment,
            tuple(episode.agents),
            tuple(episode.models),
        )
        episode_dict[episode_key] = True
    return episode_dict


def check_existing_episodes(
    env_id: str,
    agent_ids: list[str],
    models: dict[str, LLM_Name],
    index: str,
    episode_dict: dict[tuple[str, tuple[str, ...], tuple[str, ...]], bool],
) -> bool:
    models_list = (
        (models["env"], models["test_model"], models["partner_model"])
        if index == "0"
        else (models["env"], models["partner_model"], models["test_model"])
    )
    episode_key = (env_id, tuple(agent_ids), models_list)
    return episode_dict.get(episode_key, False)


def run_async_benchmark_in_batch(
    *,
    env_agent_combo_list: list[EnvAgentCombo[Observation, AgentAction]],
    batch_size: int = 1,
    tag: str = "",
    push_to_db: bool = False,
    verbose: bool = False,
) -> None:
    env_agent_combo_batch: list[EnvAgentCombo[Observation, AgentAction]] = []
    loop = asyncio.get_event_loop()
    for env_agent_combo in tqdm(
        env_agent_combo_list,
        desc="Running all envs in batch",
    ):
        env_agent_combo_batch.append(env_agent_combo)
        if len(env_agent_combo_batch) == batch_size:
            logging.info(
                f"Running batch of {batch_size} episodes: {env_agent_combo_batch}"
            )
            loop.run_until_complete(
                run_async_server(
                    sampler=BaseSampler[Observation, AgentAction](),
                    env_agent_combo_list=env_agent_combo_batch,
                    push_to_db=push_to_db,
                    tag=tag,
                )
            )
            env_agent_combo_batch = []
    else:
        if env_agent_combo_batch:
            logging.info(
                f"Running batch of {batch_size} episodes: {env_agent_combo_batch}"
            )
            loop.run_until_complete(
                run_async_server(
                    sampler=BaseSampler[Observation, AgentAction](),
                    env_agent_combo_list=env_agent_combo_batch,
                    push_to_db=push_to_db,
                    tag=tag,
                )
            )
        # remove episodes that has bad rewards
        simulated_episodes = EpisodeLog.find(EpisodeLog.tag == tag).all()
        valid_episodes = [
            not isinstance(relevant_episode.rewards[0], float)  # type: ignore
            for relevant_episode in simulated_episodes
        ]
        for valid, episode in zip(valid_episodes, simulated_episodes):
            if not valid:
                pk = episode.pk
                assert isinstance(pk, str)
                EpisodeLog.delete(pk)


def benchmark_display(
    model_list: list[str] = default_model_list,
    partner_model: str = "together_ai/meta-llama/Llama-3-70b-chat-hf",
    evaluator_model: str = "gpt-4o",
    task: str = "hard",
    output_to_jsonl: bool = False,
    save_dir: str = ".",
) -> dict[str, dict[str, tuple[float, float]]]:
    """
    Usage: sotopia benchmark-display --model-list gpt-4o --model-list together_ai/meta-llama-Llama-3-70b-chat-hf
    Aggregate all the results for the benchmark, as described in https://github.com/sotopia-lab/sotopia-space/blob/main/data_dir/models_vs_gpt35.jsonl
    """

    print(f"Displaying performance for {model_list} vs {partner_model} on task {task}")
    model_rewards_dict = dict()
    for model in model_list:
        tag = f"benchmark_{model}_{partner_model}_{evaluator_model}_{task}_trial0"
        episodes = EpisodeLog.find(EpisodeLog.tag == tag).all()
        if len(episodes) == 0:
            print(f"No episodes found for {model}")
            continue
        avg_rewards = get_avg_reward(episodes, model)  # type: ignore
        model_rewards_dict[model] = avg_rewards
        # print(f"Model: {model}, episodes: {len(episodes)}, Avg Rewards: {avg_rewards}")
        rich.print(model_rewards_dict)
    if model_rewards_dict:
        display_in_table(model_rewards_dict)
        if output_to_jsonl:
            save_to_jsonl(model_rewards_dict, partner_model, save_dir)
    else:
        print("No episodes found for any model")
    return model_rewards_dict


def _list_all_env_agent_combo_not_in_db(
    model_names: dict[str, LLM_Name],
    env_agent_combo_storage_index_list: list[tuple[EnvAgentComboStorage, str]],
    tag: str = "",
    task: str = "",
) -> list[EnvAgentCombo[Observation, AgentAction]]:
    """Iterate over each environment and return the first env-agent combo not in the database."""
    assert tag, "tag should not be empty"

    # Preprocess the episodes for fast lookup
    episode_list_with_tag: list[EpisodeLog] = EpisodeLog.find(
        EpisodeLog.tag == tag
    ).all()  # type: ignore
    episode_dict = preprocess_episode_data(episode_list_with_tag)
    list_of_env_agent_combo_storage = []
    for env_agent_combo_storage, index in env_agent_combo_storage_index_list:
        agent_ids = env_agent_combo_storage.agent_ids
        env_id = env_agent_combo_storage.env_id
        if check_existing_episodes(
            env_id=env_id,
            agent_ids=agent_ids,
            models=model_names,
            index=index,
            episode_dict=episode_dict,
        ):
            logging.info(
                f"Episode for {env_id} with agents {agent_ids} using {list(model_names.values())} already exists"
            )
            continue
        env_profile = EnvironmentProfile.get(env_id)
        env = ParallelSotopiaEnv(
            env_profile=env_profile,
            model_name=model_names["env"],
            action_order="round-robin",
            evaluators=[
                RuleBasedTerminatedEvaluator(max_turn_number=20, max_stale_turn=2),
            ],
            terminal_evaluators=[
                ReachGoalLLMEvaluator(
                    model_names["env"],
                    EvaluationForTwoAgents[SotopiaDimensions],
                ),
            ],
        )
        agent_profiles = [AgentProfile.get(id) for id in agent_ids]
        # make sure the second agent (i.e., the agent being benchmarked) is always the indexed agent
        agents = [
            LLMAgent(agent_profile=agent_profile, model_name=agent_model)
            for agent_profile, agent_model in zip(
                agent_profiles,
                [model_names["test_model"], model_names["partner_model"]]
                if index == "0"
                else [model_names["partner_model"], model_names["test_model"]],
            )
        ]
        list_of_env_agent_combo_storage.append((env, agents))
    return list_of_env_agent_combo_storage  # type: ignore


def display_in_table(
    model_rewards_dict: dict[str, dict[str, tuple[float, float]]],
) -> None:
    table = rich.table.Table(
        title="Rewards Evaluation (+/- CI bounds)",
    )
    table.add_column("Model")
    for dimension in model_rewards_dict[list(model_rewards_dict.keys())[0]].keys():
        table.add_column(dimension)
    table.add_column("Settings")
    table.add_column("Episodes")
    for model, rewards in model_rewards_dict.items():
        table.add_row(
            model,
            *(
                [f"{rewards[k][0]:.2f} ± {rewards[k][1]:.2f}" for k in rewards.keys()]
                + [f"{rewards[k][0]:.0f}" for k in ["setting_num", "episode_count"]]
            ),
        )
    rich.print(table)


def save_to_jsonl(
    model_rewards_dict: dict[str, dict[str, tuple[float, float]]],
    partner_model: str,
    save_dir: str,
) -> None:
    simplified_model_name = partner_model.split("/")[-1]
    output_fn = os.path.join(save_dir, f"models_vs_{simplified_model_name}.jsonl")
    outputs: list[str] = []
    for model, rewards in model_rewards_dict.items():
        formatted_reward = OrderedDict(
            {
                "model_name": model,
                **{
                    f"{v[0]} {v[1]}": rewards[k][0]
                    for k, v in dimension_range_mapping.items()
                },
            }
        )
        outputs.append(json.dumps(formatted_reward))
    with open(output_fn, "w") as f:
        f.write("\n".join(outputs))

    print(f"Output saved to {output_fn}")


@app.command()
def benchmark(
    models: list[str] = typer.Option(
        default_model_list,
        help=f"All the language model you want to benchmark. Default is the pre-loaded model list {default_model_list}.",
    ),
    partner_model: str = typer.Option(
        "together_ai/meta-llama/Llama-3-70b-chat-hf",
        help="The partner model you want to use.",
    ),
    evaluator_model: str = typer.Option(
        "gpt-4o", help="The evaluator model you want to use."
    ),
    batch_size: int = typer.Option(10, help="The batch size you want to use."),
    task: str = typer.Option("hard", help="The task id you want to benchmark."),
    url: str = typer.Option("", help="The url to fetch the benchmark combo."),
    print_logs: bool = typer.Option(False, help="Print logs."),
    only_show_performance: bool = typer.Option(False, help="Only show performance."),
    output_to_jsonl: bool = typer.Option(False, help="Output to jsonl."),
    push_to_db: bool = typer.Option(False, help="Push to db."),
    save_dir: str = typer.Option(".", help="The directory to save the output."),
) -> None:
    if only_show_performance:
        benchmark_display(
            models, partner_model, evaluator_model, task, output_to_jsonl, save_dir
        )
        return

    """A simple command-line interface example."""
    _set_up_logs(print_logs=print_logs)
    benchmark_combo = initialize_benchmark_combo(url)
    if task == "hard":
        hard_envs = EnvironmentList.get("01HAK34YPB1H1RWXQDASDKHSNS").environments
        agent_index = EnvironmentList.get("01HAK34YPB1H1RWXQDASDKHSNS").agent_index
        assert isinstance(agent_index, list), "agent_index should be a list"
        env_agent_combo_storage_index_list = []
        for env_id, index in zip(hard_envs, agent_index):
            for env_agent_combo_storage in benchmark_combo:
                if env_agent_combo_storage.env_id == env_id:
                    env_agent_combo_storage_index_list.append(
                        (env_agent_combo_storage, index)
                    )

    else:
        env_agent_combo_storage_index_list = [
            (env_agent_combo_storage, "1")
            for env_agent_combo_storage in benchmark_combo
        ] + [
            (env_agent_combo_storage, "0")
            for env_agent_combo_storage in benchmark_combo
        ]
    for model in models:
        typer.echo(
            f"Running benchmark for {model} chatting with {partner_model} on task {task} with {evaluator_model} as the evaluator."
        )
        model = cast(LLM_Name, model)
        partner_model = cast(LLM_Name, partner_model)
        evaluator_model = cast(LLM_Name, evaluator_model)
        tag = f"benchmark_{model}_{partner_model}_{evaluator_model}_{task}_trial0"
        typer.echo(typer.style(f"Tag: {tag}", fg=typer.colors.GREEN, bold=True))
        model_names = {
            "env": evaluator_model,
            "test_model": model,
            "partner_model": partner_model,
        }
        env_agent_combo_list = _list_all_env_agent_combo_not_in_db(
            model_names=model_names,
            tag=tag,
            env_agent_combo_storage_index_list=env_agent_combo_storage_index_list,
            task=task,
        )
        number_of_fix_turns = 0
        while True:
            run_async_benchmark_in_batch(
                env_agent_combo_list=env_agent_combo_list,
                batch_size=batch_size,
                tag=tag,
                verbose=False,
                push_to_db=push_to_db,
            )
            env_agent_combo_list = _list_all_env_agent_combo_not_in_db(
                model_names=model_names,
                tag=tag,
                env_agent_combo_storage_index_list=env_agent_combo_storage_index_list,
                task=task,
            )
            number_of_fix_turns += 1
            if len(env_agent_combo_list) == 0 or number_of_fix_turns >= 5:
                return

    benchmark_display(
        models, partner_model, evaluator_model, task, output_to_jsonl=output_to_jsonl
    )



from typing import TypeVar
from redis_om import JsonModel
from .annotators import Annotator
from .env_agent_combo_storage import EnvAgentComboStorage
from .logs import AnnotationForEpisode, EpisodeLog
from .persistent_profile import (
    AgentProfile,
    EnvironmentProfile,
    RelationshipProfile,
    RelationshipType,
)
from .serialization import (
    agentprofiles_to_csv,
    agentprofiles_to_jsonl,
    envagnetcombostorage_to_csv,
    envagnetcombostorage_to_jsonl,
    environmentprofiles_to_csv,
    environmentprofiles_to_jsonl,
    episodes_to_csv,
    episodes_to_jsonl,
    get_rewards_from_episode,
    jsonl_to_agentprofiles,
    jsonl_to_envagnetcombostorage,
    jsonl_to_environmentprofiles,
    jsonl_to_episodes,
    jsonl_to_relationshipprofiles,
    relationshipprofiles_to_csv,
    relationshipprofiles_to_jsonl,
)
from .session_transaction import MessageTransaction, SessionTransaction
from .waiting_room import MatchingInWaitingRoom
from .aggregate_annotations import map_human_annotations_to_episode_logs

__all__ = [
    "AgentProfile",
    "EnvironmentProfile",
    "EpisodeLog",
    "EnvAgentComboStorage",
    "AnnotationForEpisode",
    "Annotator",
    "RelationshipProfile",
    "RelationshipType",
    "RedisCommunicationMixin",
    "SessionTransaction",
    "MessageTransaction",
    "MatchingInWaitingRoom",
    "agentprofiles_to_csv",
    "agentprofiles_to_jsonl",
    "environmentprofiles_to_csv",
    "environmentprofiles_to_jsonl",
    "relationshipprofiles_to_csv",
    "relationshipprofiles_to_jsonl",
    "envagnetcombostorage_to_csv",
    "envagnetcombostorage_to_jsonl",
    "episodes_to_csv",
    "episodes_to_jsonl",
    "map_human_annotations_to_episode_logs",
    "jsonl_to_agentprofiles",
    "jsonl_to_environmentprofiles",
    "jsonl_to_episodes",
    "jsonl_to_relationshipprofiles",
    "jsonl_to_envagnetcombostorage",
    "get_rewards_from_episode",
]

InheritedJsonModel = TypeVar("InheritedJsonModel", bound="JsonModel")


def _json_model_all(cls: type[InheritedJsonModel]) -> list[InheritedJsonModel]:
    return cls.find().all()  # type: ignore[return-value]


JsonModel.all = classmethod(_json_model_all)  # type: ignore[assignment,method-assign]



"""XML Renderer for background, goal, observation, etc."""

from io import StringIO
from typing import cast

from beartype.door import is_bearable
from lxml import etree

from .base import BaseRenderer, RenderContext


def _render_xml(xml_node: etree._Element | str, context: RenderContext) -> str:
    if isinstance(xml_node, str):
        return xml_node
    else:
        if xml_node.tag in ["root", "p"] or xml_node.tag in context.tags_to_render:
            if context.viewer.startswith("agent_"):
                # For each agent, we only render the messages viewable by that agent
                all_visible_children = xml_node.xpath(
                    f"./node()[@viewer='{context.viewer}'] | ./node()[not(@viewer)]"
                )
                assert is_bearable(all_visible_children, list[etree._Element | str])
                cast(list[etree._Element | str], all_visible_children)
                return "".join(
                    _render_xml(child, context)
                    for child in all_visible_children  # type: ignore[attr-defined]
                )
            elif context.viewer == "human":
                # For human, we render the raw xml
                return etree.tostring(xml_node, pretty_print=True).decode("utf-8")
            elif context.viewer == "environment":
                # For environment, we render all text
                all_text = xml_node.xpath("//text()")
                return "".join(cast(list[str], all_text))
        # Add return statement for the case where none of the conditions are met
        return ""


class XMLRenderer(BaseRenderer):
    def __init__(self) -> None:
        super().__init__()
        self.parser = etree.XMLParser(recover=True, encoding="utf-8")

    def __call__(
        self, xml_string: str, context: RenderContext = RenderContext()
    ) -> str:
        if not xml_string:
            return ""
        try:
            root = etree.fromstring(xml_string)
        except etree.XMLSyntaxError:
            # try wrapping the xml_string with a pair of root tags
            try:
                root = etree.fromstring(f"<root>{xml_string}</root>")
            except etree.XMLSyntaxError:
                # try escaping the xml_string
                table = str.maketrans(
                    {
                        "&": "&amp;",
                    }
                )
                try:
                    root = etree.parse(
                        StringIO(f"<root>{xml_string.translate(table)}</root>"),
                        self.parser,
                    ).getroot()
                except etree.XMLSyntaxError as e:
                    raise etree.XMLSyntaxError(
                        f"Failed to parse xml_string: {xml_string}"
                    ) from e

        return _render_xml(root, context)



import subprocess
from typing import Literal, Optional
from pydantic import BaseModel
import rich
import rich.errors
from ..rich_pixels import Pixels

from rich.prompt import Prompt
from rich.segment import Segment
from rich.style import Style

from pathlib import Path

import typer
from .menu import Menu
import tempfile

from ..app import app


def _get_system() -> Literal["Linux", "Darwin", "Windows"]:
    import platform

    system = platform.system()
    if system == "Linux":
        return "Linux"
    elif system == "Darwin":
        return "Darwin"
    elif system == "Windows":
        return "Windows"
    else:
        raise ValueError(
            f"You are using {system} which is not supported. Please use Linux, MacOS or Windows."
        )


class Dataset(BaseModel):
    id: str
    display_name: str
    url: str
    venue: str
    license: str
    citation: str


class Datasets(BaseModel):
    datasets: list[Dataset]


def _get_dataset_to_be_loaded(
    published_datasets: Datasets, console: rich.console.Console
) -> str:
    menu = Menu(
        *(
            f"{dataset.display_name} ({dataset.venue}, {dataset.license})"
            for dataset in published_datasets.datasets
        ),
        "None of the above, I want only an empty database.",
        "No, I have a custom URL.",
        start_index=0,
        align="left",
        rule_title="Select the dataset to be loaded",
    )

    dataset = menu.ask(return_index=True)
    assert isinstance(dataset, int)

    if dataset < len(published_datasets.datasets):
        console.log(
            f"""Loading the database with data from {published_datasets.datasets[dataset].url}.
This data is from the {published_datasets.datasets[dataset].display_name}.
Licensed under {published_datasets.datasets[dataset].license}.
Please cite the following paper(s) if you use this data:
{published_datasets.datasets[dataset].citation}"""
        )
        return published_datasets.datasets[dataset].url
    elif dataset == len(published_datasets.datasets):
        console.log("Starting redis with an empty database.")
        return ""
    else:
        custom_load_database_url = Prompt.ask(
            "Enter the URL to load the database with initial data from.",
        )
        if custom_load_database_url == "":
            console.log("Starting redis with an empty database.")
            return ""
        else:
            console.log(
                f"Loading the database with initial data from {custom_load_database_url}."
            )
            return custom_load_database_url


_pixel_art = """\


███████      ███████  ██████████████████████████████████████████████████████████████████████████████████████████████████████████████████████████████████
█████        ██       ██████████████████████████████████████████████████████████████████████████████████████████████████████████████████████████████████
█████      ███        ████████         ██████████████  ███████████████████████████  ████████████████xx████████████████████████x███████████████xxx█xxx███
█████  ██████         ███████   █████   ███     ███       ████    ████  ██    ██████████     ███████xx██x██xxxxx████xxxxx███xxxxxxx███xxxxx███xxx█xxx███
█████      ███        ████████       ████    ██   ███  █████   ██   ██     █    ██  ██   ██   ██████xx██xxx██xxxx██xx████████xxx████xxx███xx██xxx█xxx███
█████        ██       █████████████     █  █████   ██  ████  █████   █  ██████  ██  ██  ████  ██████xx██xx█████xx██xxxxxx████xxx████xx████xx██xxx█xxx███
█████         ██████  ███████   ██████  █   ████  ███  ████   ████   █   ████   ██  █   ████  ██████xx██xx█████xx████████xx██xxx████xx████xx██xxx█xxx███
█████        ███      █████████        ███       █████    ███      ███  █      ███  ██         █████xx██xx█████xx██xxxxxxxx███xxxxx█xxxxxxxxx█xxx█xxx███
█████       ██        ████████████████████████████████████████████████  ████████████████████████████████████████████████████████████████████████████████
█████   █████       ██████████████████████████████████████████████████  ████████████████████████████████████████████████████████████████████████████████


"""


@app.command()
def install(
    use_docker: Optional[bool] = typer.Option(None, help="Install redis using docker."),
    load_database: Optional[bool] = typer.Option(
        None, help="Load the database with initial data."
    ),
    load_sotopia_pi_data: bool = typer.Option(
        True,
        help="Load the database with initial data from Sotopia π. Only applicable if `load_database` is True.",
    ),
    custom_database_url: Optional[str] = typer.Option(
        None, help="Load the database with initial data from a custom URL."
    ),
    redis_data_path: Optional[str] = typer.Option(
        None,
        help="Path to store the redis data. Only applicable if `use_docker` is True.",
    ),
    overwrite_existing_data: Optional[bool] = typer.Option(
        None, help="Overwrite existing data in the redis data path."
    ),
) -> None:
    console = rich.console.Console()
    mapping = {
        " ": Segment(" ", Style.parse("black on black")),
        "█": Segment("█", Style.parse("white")),
        "x": Segment("x", Style.parse("magenta on magenta")),
    }
    pixels = Pixels.from_ascii(_pixel_art, mapping)
    console.print(pixels, justify="center")
    system = _get_system()

    if use_docker is None:
        if system == "Windows":
            console.log(
                "Installing Redis with Docker... Check if Docker Desktop is installed."
            )
            use_docker = True
        elif system == "Darwin":
            use_docker = (
                Prompt.ask(
                    "Do you want to use Docker or Homebrew to install redis? We recommand you to use Docker.",
                    choices=["Docker", "Homebrew"],
                    default="Docker",
                    console=console,
                )
                == "Docker"
            )
        else:
            use_docker = (
                Prompt.ask(
                    "Do you want to use Docker to install redis or directly install Redis stack binary? We recommand you to use Docker.",
                    choices=["Docker", "Binary"],
                    default="Docker",
                    console=console,
                )
                == "Docker"
            )

    if use_docker:
        try:
            subprocess.check_output("command -v docker", shell=True)
        except subprocess.CalledProcessError:
            if system == "Darwin":
                console.log(
                    """Docker is not installed.
                    Please check https://docs.docker.com/desktop/install/mac-install/,
                    or install it using homebrew with `brew install --cask docker`.
                    And then run this command again.
                    """
                )
            elif system == "Linux":
                console.log(
                    """Docker is not installed.
                    Please check https://docs.docker.com/engine/install/ubuntu/.
                    And then run this command again.
                    """
                )
            else:
                console.log(
                    """Docker is not installed.
                    Please check https://docs.docker.com/desktop/install/windows-install/.
                    And then run this command again.
                    """
                )
            exit(1)
        try:
            subprocess.check_output("docker ps", shell=True)
        except subprocess.CalledProcessError:
            if system == "Darwin":
                console.log(
                    """Docker Daemon was not started. Please launch Docker App.
                    """
                )
            elif system == "Linux":
                console.log(
                    """Docker Daemon was not started. Please run `dockerd`
                    """
                )
            else:
                console.log(
                    """Docker Daemon was not started. Please launch Docker App.
                    """
                )
            exit(1)
    else:
        if system == "Windows":
            console.log("""For Windows, unfortunately only docker is supported.
                Check the official documentation:
                https://redis.io/docs/latest/operate/oss_and_stack/install/install-stack/windows/.
            """)
            exit(1)
        elif system == "Darwin":
            # check if homebrew is installed
            try:
                subprocess.check_output("command -v brew", shell=True)
                subprocess.run("brew update-reset", shell=True, check=True)
            except subprocess.CalledProcessError:
                console.log(
                    """Homebrew is required for install redis without docker on MacOS.
                    Please check https://brew.sh/.
                    And then run this command again.
                    """
                )
                exit(1)

    next_state: Literal["ask_data_source", "empty_database", "custom_url"] | None = None
    url = ""

    if load_database is None:
        load_database = (
            Prompt.ask(
                "Do you want to load the database with published data?",
                choices=["Yes", "No"],
                default="Yes",
                console=console,
            )
            == "Yes"
        )
        if load_database:
            next_state = "ask_data_source"
        else:
            next_state = "empty_database"
    elif not load_database:
        next_state = "empty_database"
    else:
        if load_sotopia_pi_data:
            next_state = "custom_url"
            url = "https://huggingface.co/datasets/cmu-lti/sotopia-pi/resolve/main/dump.rdb?download=true"
        elif custom_database_url is None:
            next_state = "ask_data_source"
        else:
            next_state = "custom_url"
            url = custom_database_url

    if next_state == "ask_data_source":
        fn = Path(__file__).parent / "published_datasets.json"
        published_datasets = Datasets.parse_file(fn)
        url = _get_dataset_to_be_loaded(published_datasets, console)
        next_state = "custom_url"

    assert next_state in ["custom_url", "empty_database"]

    tmpdir_context = tempfile.TemporaryDirectory()
    tmpdir = tmpdir_context.__enter__()

    if url:
        try:
            subprocess.run(f"curl -L {url} -o {Path(tmpdir) / 'dump.rdb'}", shell=True)
            console.log("Database downloaded successfully.")
        except subprocess.CalledProcessError:
            console.log("Database download failed. Please check the URL and try again.")
    else:
        console.log("Starting redis with an empty database.")

    if use_docker:
        current_directory = Path(__file__).parent
        if redis_data_path is None:
            directory = Prompt.ask(
                "Enter the directory where you want to store the data. Press enter to use the current directory.",
                default=current_directory,
            )
        else:
            directory = redis_data_path
        (Path(directory) / "redis-data").mkdir(parents=True, exist_ok=True)
        if load_database:
            if Path.exists(Path(directory) / "redis-data" / "dump.rdb"):
                cover_existing = (
                    Prompt.ask(
                        "The directory already contains a dump.rdb file. Do you want to overwrite it?",
                        choices=["Yes", "No"],
                        default="No",
                        console=console,
                    )
                    if overwrite_existing_data is None
                    else "Yes"
                    if overwrite_existing_data
                    else "No"
                )
                if cover_existing == "No":
                    console.log(
                        "Exiting the installation. Please provide a different directory."
                    )
                    exit(0)
            else:
                (Path(tmpdir) / "dump.rdb").rename(
                    Path(directory) / "redis-data/dump.rdb"
                )
        try:
            subprocess.run(
                f"docker run -d --name redis-stack -p 6379:6379 -p 8001:8001 -v {directory}/redis-data:/data/ redis/redis-stack:latest",
                shell=True,
                check=True,
            )
            console.log(
                "Redis started successfully. To stop redis, run `docker stop redis-stack`. To restart, run `docker restart redis-stack`."
            )
        except subprocess.CalledProcessError:
            console.log("Redis start failed. Please check the logs and try again.")
    else:
        if system == "Darwin":
            try:
                subprocess.run(
                    "brew tap redis-stack/redis-stack", shell=True, check=True
                )
                subprocess.run("brew install redis-stack", shell=True, check=True)
                if load_database:
                    if Path("/opt/homebrew/var/db/redis-stack/dump.rdb").exists():
                        cover_existing = (
                            Prompt.ask(
                                "The directory already contains a dump.rdb file. Do you want to overwrite it?",
                                choices=["Yes", "No"],
                                default="No",
                                console=console,
                            )
                            if overwrite_existing_data is None
                            else "Yes"
                            if overwrite_existing_data
                            else "No"
                        )
                        if cover_existing == "No":
                            console.log(
                                "Exiting the installation. Please provide a different directory."
                            )
                            exit(0)
                    else:
                        Path("/opt/homebrew/var/db/redis-stack/").mkdir(
                            parents=True, exist_ok=True
                        )
                    subprocess.run(
                        f"mv {tmpdir}/dump.rdb /opt/homebrew/var/db/redis-stack/dump.rdb",
                        shell=True,
                        check=True,
                    )
                subprocess.run(
                    "redis-stack-server --daemonize yes", shell=True, check=True
                )
                console.log(
                    "Redis started successfully. To stop redis, run `redis-cli shutdown`. To restart, run this script again."
                )
            except subprocess.CalledProcessError:
                console.log("Redis start failed. Please check the logs and try again.")
        elif system == "Linux":
            try:
                subprocess.run(
                    "curl -fsSL https://packages.redis.io/redis-stack/redis-stack-server-7.2.0-v10.focal.x86_64.tar.gz -o redis-stack-server.tar.gz",
                    shell=True,
                    check=True,
                )
                subprocess.run(
                    "tar -xvzf redis-stack-server.tar.gz", shell=True, check=True
                )
                if load_database:
                    Path("./redis-stack-server-7.2.0-v10/var/db/redis-stack").mkdir(
                        parents=True, exist_ok=True
                    )
                    subprocess.run(
                        f"mv {tmpdir}/dump.rdb ./redis-stack-server-7.2.0-v10/var/db/redis-stack/dump.rdb",
                        shell=True,
                        check=True,
                    )
                subprocess.run(
                    "./redis-stack-server-7.2.0-v10/bin/redis-stack-server --daemonize yes",
                    shell=True,
                    check=True,
                )
                console.log(
                    "Redis started successfully. To stop redis, run `./redis-stack-server-7.2.0-v10/bin/redis-cli shutdown`. To restart, run this script again."
                )
            except subprocess.CalledProcessError:
                console.log("Redis start failed. Please check the logs and try again.")

    tmpdir_context.__exit__(None, None, None)


if __name__ == "__main__":
    app()



import json
import os
from typing import Callable

from sotopia.generation_utils.generate import convert_narratives, agenerate_init_profile
from sotopia.messages import Message, ScriptBackground


async def generate_background(
    info_json_file: str, basic_info: dict[str, str]
) -> tuple[str, str, str, str, list[dict[str, str]]]:
    if os.path.isfile(info_json_file):
        with open(info_json_file, "r") as f:
            info_dict = json.load(f)
            initial_profile = str(info_dict["initial_profile"])
            profile = info_dict["profile"]
            first_narrative = info_dict["first_narrative_profile"]
            second_narrative = info_dict["second_narrative_profile"]
            previous_messages = info_dict["messages"]
    else:
        initial_profile = str(basic_info)
        profile = await agenerate_init_profile(
            model_name="gpt-4o-mini", basic_info=basic_info
        )
        first_narrative = convert_narratives(
            model_name="gpt-4o-mini", narrative="first", text=profile
        )
        second_narrative = convert_narratives(
            model_name="gpt-4o-mini", narrative="second", text=profile
        )
        previous_messages = []
    return (
        initial_profile,
        profile,
        first_narrative,
        second_narrative,
        previous_messages,
    )


def generate_background_conversation(
    seed: dict[str, str],
    basic_info: dict[str, str],
    initial_profile: str,
    profile: str,
    background_json_file: str,
    run_sync_server: Callable[..., list[tuple[str, str, Message]]],
) -> tuple[list[tuple[str, str, Message]], ScriptBackground]:
    scenario, _topic, role, q_goal, a_goal = (
        seed["scenario"],
        seed["topic"],
        seed["role"],
        seed["q_goal"],
        seed["a_goal"],
    )
    background = ScriptBackground(
        scenario=scenario,
        p1_name=role,
        p2_name=basic_info["name"],
        p1_background=role,
        p2_background=initial_profile + "\n" + profile,
        p1_goal=q_goal,
        p2_goal=a_goal,
    )
    with open(background_json_file, "w") as f:
        background_dict = json.loads(background.json())
        json.dump(background_dict, f, indent=4)

    model_names: dict[str, str] = {
        "env": "gpt-4o-mini",
        "agent2": "gpt-4o-mini",
        "agent1": "gpt-4",
    }

    agents_info: dict[str, dict[str, str]] = {
        "env": {"mode": "all"},
        basic_info["name"]: {
            "mode": "speak",
            "goal": background.p2_goal,
        },
        role: {
            "mode": "speak",
            "goal": background.p1_goal,
        },
    }

    messages = run_sync_server(
        model_name_dict=model_names,
        agents_info=agents_info,
        action_order="round-robin",
        full_background_file=background_json_file,
        mode="speak",
    )
    return messages, background



from sotopia.database import EpisodeLog, AnnotationForEpisode
from typing import List, Tuple, Dict
from copy import deepcopy
from collections import defaultdict


def aggregate_reasoning(reasonings: List[str]) -> str:
    return "\n".join(reasonings)


def aggregate_rewards(
    rewards: List[Tuple[float, Dict[str, float]]],
) -> Tuple[float, Dict[str, float]]:
    def average_dict(dict_list: List[Dict[str, float]]) -> Dict[str, float]:
        keys = dict_list[0].keys()
        return {k: sum(d[k] for d in dict_list) / len(dict_list) for k in keys}

    def average_list(item: List[float]) -> float:
        return sum(item) / len(item)

    ret_rewards = (
        average_list([r[0] for r in rewards]),
        average_dict([r[1] for r in rewards]),
    )
    return ret_rewards


def map_human_annotations_to_episode_logs(
    human_annotation: list[AnnotationForEpisode],
    return_model_episodes: bool = False,
    aggregate: bool = False,
) -> dict[str, EpisodeLog | tuple[EpisodeLog, EpisodeLog]]:
    """
    Map human annotations to corresponding episode logs.

    Args:
        human_annotations (List[AnnotationForEpisode]): List of annotations for episodes.
        return_model_episodes (bool, optional): If True, returns a tuple of human and model episodes. Defaults to False.
        aggregate (bool, optional): If True, aggregates reasoning and rewards across all human annotations per model episode. The primary key will be the model episode. Defaults to False.

    Returns:
        Dict[str, Union[EpisodeLog, Tuple[EpisodeLog, EpisodeLog]]]: A dictionary mapping episode primary keys to EpisodeLog objects or tuples of (human_episode, model_episode) depending on
        return_model_episodes and aggregate flags.
    """

    model_human_pk_mapping: Dict[str, List[str]] = defaultdict(list)
    for annotation in human_annotation:
        model_episode_pk = annotation.episode
        human_episode_pk = annotation.pk
        assert human_episode_pk is not None
        model_human_pk_mapping[model_episode_pk].append(human_episode_pk)

    ep_dict: Dict[str, EpisodeLog | Tuple[EpisodeLog, EpisodeLog]] = {}

    if aggregate:
        for model_episode_pk, human_episode_pks in model_human_pk_mapping.items():
            all_human_rewards = []
            all_human_reasonings = []

            for human_episode_pk in human_episode_pks:
                annotation = AnnotationForEpisode.get(pk=human_episode_pk)

                human_reasoning = annotation.reasoning
                human_rewards = annotation.rewards

                if any(isinstance(r, float) for r in human_rewards):
                    print("Rewards are not in the correct format")
                    print(human_rewards)
                    continue

                all_human_rewards.append(human_rewards)
                all_human_reasonings.append(human_reasoning)

            episode = EpisodeLog.get(pk=model_episode_pk)
            human_episode = deepcopy(episode)
            human_reasoning = aggregate_reasoning(all_human_reasonings)

            human_rewards = [
                aggregate_rewards([r[0] for r in all_human_rewards]),  # type: ignore
                aggregate_rewards([r[1] for r in all_human_rewards]),  # type: ignore
            ]
            if human_reasoning:
                human_episode.reasoning = human_reasoning
            if human_rewards:
                human_episode.rewards = human_rewards

            if return_model_episodes:
                ep_dict[model_episode_pk] = (human_episode, episode)
            else:
                ep_dict[model_episode_pk] = human_episode
        return ep_dict

    for annotation in human_annotation:
        episode_pk = annotation.episode
        episode = EpisodeLog.get(pk=episode_pk)
        human_episode_pk = annotation.pk
        assert human_episode_pk is not None

        human_reasoning = annotation.reasoning
        human_rewards = annotation.rewards

        human_episode = deepcopy(episode)
        human_episode.reasoning = human_reasoning
        human_episode.rewards = human_rewards

        if return_model_episodes:
            ep_dict[human_episode_pk] = (human_episode, episode)
        else:
            ep_dict[human_episode_pk] = human_episode

    return ep_dict



from .base_sampler import BaseSampler, EnvAgentCombo
from .constraint_based_sampler import ConstraintBasedSampler
from .uniform_sampler import UniformSampler

__all__ = [
    "BaseSampler",
    "UniformSampler",
    "ConstraintBasedSampler",
    "EnvAgentCombo",
]



from redis_om import JsonModel
from redis_om.model.model import Field

DEFAULT_EXPIRE_TIME = 60 * 60 * 24 * 7  # 7 day


class AutoExpireMixin(JsonModel):
    expire_time: int = Field(default=DEFAULT_EXPIRE_TIME, index=True)

    def save(self) -> None:
        super().save()
        self.expire(self.expire_time)



import asyncio
from concurrent.futures import ThreadPoolExecutor
from functools import wraps
from .generate import agenerate, agenerate_action

from typing import Awaitable, Callable, ParamSpec, TypeVar

P = ParamSpec("P")
T = TypeVar("T")


def async_to_sync(async_func: Callable[P, Awaitable[T]]) -> Callable[P, T]:
    @wraps(async_func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        with ThreadPoolExecutor() as executor:
            future = executor.submit(
                loop.run_until_complete, async_func(*args, **kwargs)
            )
            result = future.result()
        loop.close()
        return result

    return wrapper


generate = async_to_sync(agenerate)
generate_action = async_to_sync(agenerate_action)



from pydantic import BaseModel, Field, validator


class RenderContext(BaseModel):
    viewer: str = Field(
        default="human",
        description="The viewer of the rendered string."
        "Must be one of ['human', 'environment', 'agent_i']"
        "where i is a non-negative integer, the index of the agent in the episode log.",
    )

    verbose: bool = Field(
        default=False,
        description="Whether to render the verbose version of the string.",
    )
    tags_to_render: list[str] = Field(
        default=[], description="The special tags to render."
    )

    @validator("viewer")
    def viewer_must_be_valid(cls, v: str) -> str:
        if v.startswith("agent_"):
            try:
                agent_idx = int(v.split("_")[1])
                assert agent_idx >= 0
            except Exception:
                raise ValueError(
                    "If viewer is an agent, it must be of the form agent_i, where i is a non-negative integer."
                )
        elif v not in ["human", "environment"]:
            raise ValueError(
                "Viewer must be one of ['human', 'environment', 'agent_i']"
                "where i is a non-negative integer, the index of the agent in the episode log."
            )
        return v


class BaseRenderer(object):
    def __call__(self, input_string: str, context: RenderContext) -> str:
        return input_string



from redis_om import JsonModel
from redis_om.model.model import Field

from .auto_expires_mixin import AutoExpireMixin


class MatchingInWaitingRoom(AutoExpireMixin, JsonModel):
    timestamp: float = Field()
    client_ids: list[str] = Field(default_factory=lambda: [])
    session_ids: list[str] = Field(default_factory=lambda: [])
    session_id_retrieved: list[str] = Field(default_factory=lambda: [])



from .message_classes import Message


class MessengerMixin(object):
    def __init__(self) -> None:
        self.inbox: list[tuple[str, Message]] = []

    def reset_inbox(self) -> None:
        self.inbox = []

    def recv_message(self, source: str, message: Message) -> None:
        self.inbox.append((source, message))



from __future__ import annotations

from typing import Iterable, Mapping, Optional

from rich.console import Console, ConsoleOptions, RenderResult
from rich.segment import Segment, Segments


class Pixels:
    def __init__(self) -> None:
        self._segments: Segments | None = None

    @staticmethod
    def from_segments(
        segments: Iterable[Segment],
    ) -> Pixels:
        """Create a Pixels object from an Iterable of Segments instance."""
        pixels = Pixels()
        pixels._segments = Segments(segments)
        return pixels

    @staticmethod
    def from_ascii(
        grid: str, mapping: Optional[Mapping[str, Segment]] = None
    ) -> Pixels:
        """
        Create a Pixels object from a 2D-grid of ASCII characters.
        Each ASCII character can be mapped to a Segment (a character and style combo),
        allowing you to add a splash of colour to your grid.

        Args:
            grid: A 2D grid of characters (a multi-line string).
            mapping: Maps ASCII characters to Segments. Occurrences of a character
                will be replaced with the corresponding Segment.
        """
        if mapping is None:
            mapping = {}

        if not grid:
            return Pixels.from_segments([])

        segments = []
        for character in grid:
            segment = mapping.get(character, Segment(character))
            segments.append(segment)

        return Pixels.from_segments(segments)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield self._segments or ""



import ast
import random
from typing import Any, Generator, Type, TypeVar

from sotopia.agents.base_agent import BaseAgent
from sotopia.database import (
    AgentProfile,
    EnvironmentProfile,
    RelationshipProfile,
)
from sotopia.envs.parallel import ParallelSotopiaEnv

from .base_sampler import BaseSampler, EnvAgentCombo

ObsType = TypeVar("ObsType")
ActType = TypeVar("ActType")


def _get_fit_agents_for_one_env(
    env_profile_id: str, agent_candidate_ids: set[str] | None, size: int
) -> list[list[str]]:
    env = EnvironmentProfile.get(env_profile_id)

    relationship_constraint = env.relationship
    available_relationships = RelationshipProfile.find(
        RelationshipProfile.relationship == relationship_constraint
    ).all()
    age_contraint = env.age_constraint
    assert isinstance(age_contraint, str)
    if age_contraint != "[(18, 70), (18, 70)]":
        age_contraint_list = ast.literal_eval(age_contraint)
        available_relationships = [
            relationship
            for relationship in available_relationships
            if (
                age_contraint_list[0][0]
                <= AgentProfile.get(relationship.agent_1_id).age  # type: ignore[attr-defined]
                <= age_contraint_list[0][1]
                and age_contraint_list[1][0]
                <= AgentProfile.get(relationship.agent_2_id).age  # type: ignore[attr-defined]
                <= age_contraint_list[1][1]
            )
        ]
    if len(available_relationships) < size:
        raise ValueError(
            f"Number of available relationships ({len(available_relationships)}) "
            f"is smaller than the required size ({size})"
        )
    random.shuffle(available_relationships)
    selected_relationship = available_relationships[:size]
    fit_agents = []
    for relationship in selected_relationship:
        assert isinstance(relationship, RelationshipProfile)
        fit_agents.append([relationship.agent_1_id, relationship.agent_2_id])
    return fit_agents


class ConstraintBasedSampler(BaseSampler[ObsType, ActType]):
    def sample(
        self,
        agent_classes: Type[BaseAgent[ObsType, ActType]]
        | list[Type[BaseAgent[ObsType, ActType]]],
        n_agent: int = 2,
        replacement: bool = True,
        size: int = 5,
        env_params: dict[str, Any] = {},
        agents_params: list[dict[str, Any]] = [{}, {}],
    ) -> Generator[EnvAgentCombo[ObsType, ActType], None, None]:
        """
        Sample an environment and a list of agents based on the constraints of the environment.

        Note: Sampling without replacement is only restricted to single env candidate.
        This is due to the fact that the number of possible combinations of env and agents is huge.
        Please sample for each env separately if you want to sample without replacement.
        """
        assert (
            not isinstance(agent_classes, list) or len(agent_classes) == n_agent
        ), f"agent_classes should be a list of length {n_agent} or a single agent class"

        if not isinstance(agent_classes, list):
            agent_classes = [agent_classes] * n_agent
        assert (
            len(agents_params) == n_agent
        ), f"agents_params should be a list of length {n_agent}"

        env_profiles: list[EnvironmentProfile] = []
        agents_which_fit_scenario: list[list[str]] = []

        if self.env_candidates is None:
            env_candidates = EnvironmentProfile.all()
            if not env_candidates:
                raise ValueError("No environment candidates available for sampling.")
            self.env_candidates = env_candidates

        if self.agent_candidates is None:
            agent_candidates = AgentProfile.all()
            if not agent_candidates:
                raise ValueError("No agent candidates available for sampling.")
            self.agent_candidates = agent_candidates

        agent_candidate_ids: set[str] | None = None
        if self.agent_candidates:
            agent_candidate_ids = set(
                str(agent.pk) if not isinstance(agent, str) else agent
                for agent in self.agent_candidates
            )
        else:
            agent_candidate_ids = None

        if not replacement:
            assert self.env_candidates and len(self.env_candidates) == 1, (
                "Sampling without replacement is only restricted to single env candidate (must be provided in the constructor). "
                "This is due to the fact that the number of possible combinations of env and agents is huge. "
                "Please sample for each env separately if you want to sample without replacement."
            )

            env_profile_id = (
                self.env_candidates[0].pk
                if not isinstance(self.env_candidates[0], str)
                else self.env_candidates[0]
            )

            assert env_profile_id, "Env candidate must have an id"

            agents_which_fit_scenario = _get_fit_agents_for_one_env(
                env_profile_id, agent_candidate_ids, size
            )
            env_profiles = (
                [EnvironmentProfile.get(env_profile_id)] * size
                if isinstance(self.env_candidates[0], str)
                else [self.env_candidates[0]] * size
            )
        else:
            for _ in range(size):
                env_profile = random.choice(self.env_candidates)
                if isinstance(env_profile, str):
                    env_profile = EnvironmentProfile.get(env_profile)
                env_profiles.append(env_profile)
                env_profile_id = env_profile.pk
                assert env_profile_id, "Env candidate must have an id"
                agents_which_fit_scenario.append(
                    _get_fit_agents_for_one_env(env_profile_id, agent_candidate_ids, 1)[
                        0
                    ]
                )

        assert len(env_profiles) == size, "Number of env_profiles is not equal to size"
        assert (
            len(agents_which_fit_scenario) == size
        ), "Number of agents_which_fit_scenario is not equal to size"

        for env_profile, agent_profile_id_list in zip(
            env_profiles, agents_which_fit_scenario
        ):
            env = ParallelSotopiaEnv(env_profile=env_profile, **env_params)
            agent_profiles = [AgentProfile.get(id) for id in agent_profile_id_list]

            agents = [
                agent_class(agent_profile=agent_profile, **agent_params)
                for agent_class, agent_profile, agent_params in zip(
                    agent_classes, agent_profiles, agents_params
                )
            ]
            # set goal for each agent
            for agent, goal in zip(agents, env.profile.agent_goals):
                agent.goal = goal

            yield env, agents



from .base_agent import BaseAgent
from .generate_agent_background import (
    generate_background,
    generate_background_conversation,
)
from .llm_agent import (
    Agents,
    HumanAgent,
    LLMAgent,
    ScriptWritingAgent,
)
from .redis_agent import RedisAgent

__all__ = [
    "BaseAgent",
    "LLMAgent",
    "Agents",
    "HumanAgent",
    "generate_background",
    "generate_background_conversation",
    "RedisAgent",
    "ScriptWritingAgent",
]



import logging
from typing import Any
from langchain_core.messages import BaseMessage
from langchain.callbacks import StdOutCallbackHandler

logging.addLevelName(15, "LangChain")


class LoggingCallbackHandler(StdOutCallbackHandler):
    """Callback Handler that prints to std out."""

    always_verbose = True

    def __init__(self, name: str) -> None:
        """Initialize callback handler."""
        super().__init__()
        self.logger = logging.getLogger(name)
        self.prompt = ""

    def on_chat_model_start(
        self,
        serialized: dict[str, Any],
        messages: list[list[BaseMessage]],
        **kwargs: Any,
    ) -> None:
        self.prompt = str(messages[0][0].content)
        logging.log(15, f"Prompt after formatting:{self.prompt}")

    def on_chain_start(self, *args: Any, **kwargs: Any) -> None:
        pass

    def on_chain_end(self, *args: Any, **kwargs: Any) -> None:
        pass

    def on_agent_action(self, *args: Any, **kwargs: Any) -> Any:
        pass

    def on_tool_end(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        pass

    def on_tool_error(
        self, error: BaseException | KeyboardInterrupt, **kwargs: Any
    ) -> None:
        """Do nothing."""
        pass

    def on_text(
        self,
        text: str,
        color: str | None = None,
        end: str = "",
        **kwargs: Any,
    ) -> None:
        """Run when agent ends."""
        # leave only prompt for environment
        pass

    def retrive_prompt(self) -> str:
        return self.prompt

    def on_agent_finish(self, *args: Any, **kwargs: Any) -> None:
        """Run on agent end."""
        pass



import random
from typing import Any, Generator, Type, TypeVar

from sotopia.agents.base_agent import BaseAgent
from sotopia.database import AgentProfile, EnvironmentProfile
from sotopia.envs.parallel import ParallelSotopiaEnv

from .base_sampler import BaseSampler, EnvAgentCombo

ObsType = TypeVar("ObsType")
ActType = TypeVar("ActType")


class UniformSampler(BaseSampler[ObsType, ActType]):
    def sample(
        self,
        agent_classes: Type[BaseAgent[ObsType, ActType]]
        | list[Type[BaseAgent[ObsType, ActType]]],
        n_agent: int = 2,
        replacement: bool = True,
        size: int = 1,
        env_params: dict[str, Any] = {},
        agents_params: list[dict[str, Any]] = [{}, {}],
    ) -> Generator[EnvAgentCombo[ObsType, ActType], None, None]:
        """
        Sample an environment and `n_agent` agents.

        Runtime checks:
        1. If `agent_classes` is a list, it should have length `n_agent`.
        2. `agents_params` should also be a list of length `n_agent`.

        Note: Currently, uniform sampling without replacement is not supported.
        This is due to the difficulty of sequentially sampling environment and agents.
        In theory, we can reject samples that have been sampled before, but this is not efficient.
        Please open an issue if you need this feature.
        """
        assert (
            not isinstance(agent_classes, list) or len(agent_classes) == n_agent
        ), f"agent_classes should be a list of length {n_agent} or a single agent class"

        if not isinstance(agent_classes, list):
            agent_classes = [agent_classes] * n_agent
        assert (
            len(agents_params) == n_agent
        ), f"agents_params should be a list of length {n_agent}"

        assert replacement, "Uniform sampling without replacement is not supported yet"

        if self.env_candidates is None:
            env_candidates = EnvironmentProfile.all()
            if not env_candidates:
                raise ValueError("No environment candidates available for sampling.")
            self.env_candidates = env_candidates

        if self.agent_candidates is None:
            agent_candidates = AgentProfile.all()
            if not agent_candidates:
                raise ValueError("No agent candidates available for sampling.")
            self.agent_candidates = agent_candidates

        for _ in range(size):
            env_profile = random.choice(self.env_candidates)
            if isinstance(env_profile, str):
                env_profile = EnvironmentProfile.get(env_profile)
            env = ParallelSotopiaEnv(env_profile=env_profile, **env_params)

            agent_profile_candidates = self.agent_candidates
            if len(agent_profile_candidates) == n_agent:
                agent_profiles_maybe_id = agent_profile_candidates
            else:
                agent_profiles_maybe_id = random.sample(
                    agent_profile_candidates, n_agent
                )
            agent_profiles = [
                i if isinstance(i, AgentProfile) else AgentProfile.get(i)
                for i in agent_profiles_maybe_id
            ]
            agents = [
                agent_class(agent_profile=agent_profile, **agent_params)
                for agent_class, agent_profile, agent_params in zip(
                    agent_classes, agent_profiles, agents_params
                )
            ]
            # set goal for each agent
            for agent, goal in zip(agents, env.profile.agent_goals):
                agent.goal = goal

            yield env, agents



import asyncio
import copy
import itertools
import random
from typing import Any, Literal, Optional, Type, TypeVar

from beartype import beartype
from gin import configurable
from gymnasium.spaces.dict import Dict
from gymnasium.spaces.discrete import Discrete
from gymnasium.spaces.text import Text
from pettingzoo.utils.env import ParallelEnv
from redis_om.model.model import NotFoundError

from sotopia.agents.llm_agent import Agents
from sotopia.database import EnvironmentProfile
from sotopia.database.persistent_profile import (
    AgentProfile,
    RelationshipType,
)
from sotopia.messages import (
    ActionType,
    AgentAction,
    MessengerMixin,
    Observation,
    ScriptBackground,
    SimpleMessage,
)
from sotopia.renderers import RenderContext, XMLRenderer

from .evaluators import Evaluator, unweighted_aggregate_evaluate

TBackground = TypeVar("TBackground", bound=ScriptBackground)


def _actions_to_natural_language(actions: dict[str, AgentAction]) -> str:
    action_str = ""
    for agent, action in actions.items():
        # Only record actions that did something
        if action.action_type != "none":
            if action_str != "":
                action_str += ";"  # separate actions with semicolon
            action_str += f"{agent} {action.to_natural_language()}"
    return action_str


def _map_gender_to_adj(gender: str) -> str:
    gender_to_adj = {
        "Man": "male",
        "Woman": "female",
        "Nonbinary": "nonbinary",
    }
    if gender:
        return gender_to_adj[gender]
    else:
        return ""


def _agent_profile_to_stranger_self(profile: AgentProfile, agent_id: int) -> str:
    return f"<root><p viewer='agent_{agent_id}'>{profile.first_name} {profile.last_name} is a {profile.age}-year-old {_map_gender_to_adj(profile.gender)} {profile.occupation.lower()}. {profile.gender_pronoun} pronouns. {profile.public_info} Personality and values description: {profile.personality_and_values} {profile.first_name}'s secrets: {profile.secret}</p></root>"


def _agent_profile_to_name_self(profile: AgentProfile, agent_id: int) -> str:
    return f"{profile.first_name} {profile.last_name} <p viewer='agent_{agent_id}'>is a {profile.age}-year-old {_map_gender_to_adj(profile.gender)} {profile.occupation.lower()}. {profile.gender_pronoun} pronouns. {profile.public_info} Personality and values description: {profile.personality_and_values} {profile.first_name}'s secrets: {profile.secret}</p>"


def _agent_profile_to_aquaintance_self(profile: AgentProfile, agent_id: int) -> str:
    return f"{profile.first_name} {profile.last_name} is a {profile.age}-year-old {_map_gender_to_adj(profile.gender)} {profile.occupation.lower()}. {profile.gender_pronoun} pronouns. {profile.public_info} <p viewer='agent_{agent_id}'>Personality and values description: {profile.personality_and_values} {profile.first_name}'s secrets: {profile.secret}</p>"


def _agent_profile_to_friendabove_self(profile: AgentProfile, agent_id: int) -> str:
    return f"{profile.first_name} {profile.last_name} is a {profile.age}-year-old {_map_gender_to_adj(profile.gender)} {profile.occupation.lower()}. {profile.gender_pronoun} pronouns. {profile.public_info} Personality and values description: {profile.personality_and_values} <p viewer='agent_{agent_id}'>{profile.first_name}'s secrets: {profile.secret}</p>"


def get_bio(
    relationship: RelationshipType, profile: AgentProfile, agent_id: int
) -> str:
    match relationship:
        case RelationshipType.stranger:
            return _agent_profile_to_stranger_self(profile, agent_id=agent_id)
        case RelationshipType.know_by_name:
            return _agent_profile_to_name_self(profile, agent_id=agent_id)
        case RelationshipType.acquaintance:
            return _agent_profile_to_aquaintance_self(profile, agent_id=agent_id)
        case (
            RelationshipType.friend
            | RelationshipType.romantic_relationship
            | RelationshipType.family_member
        ):
            return _agent_profile_to_friendabove_self(profile, agent_id=agent_id)
        case _:
            raise ValueError(f"Unknown relationship {relationship}")


@configurable
def render_text_for_agent(
    raw_text: str,
    agent_id: int,
    tags_to_render: list[str] = [
        "extra_info",
        "clarification_hint",
        "strategy_hint",
    ],
) -> str:
    return XMLRenderer()(
        raw_text,
        RenderContext(viewer=f"agent_{agent_id}", tags_to_render=tags_to_render),
    )


@configurable
def render_text_for_environment(
    raw_text: str,
    tags_to_render: list[str] = [
        "extra_info",
        "clarification_hint",
        "strategy_hint",
    ],
) -> str:
    return XMLRenderer()(
        raw_text,
        RenderContext(viewer="environment", tags_to_render=tags_to_render),
    )


class ParallelSotopiaEnv(ParallelEnv[str, Observation, AgentAction], MessengerMixin):
    def __init__(
        self,
        available_action_types: set[ActionType] = set(
            ["none", "speak", "non-verbal communication", "action", "leave"]
        ),
        action_order: Literal["simultaneous", "round-robin", "random"] = "simultaneous",
        model_name: str = "gpt-4o-mini",
        evaluators: list[Evaluator] = [],
        terminal_evaluators: list[Evaluator] = [],
        uuid_str: str | None = None,
        env_profile: EnvironmentProfile | None = None,
        background_class: Optional[Type[TBackground]] = None,
    ) -> None:
        """A sotopia environment for parallel agents.

        Args:
            available_action_types (set[ActionType], optional): The action types that are available to the agents. Defaults to set(["none", "speak", "non-verbal communication", "action"]).
            action_order (Literal["simultaneous", "round-robin", "random"], optional): The order in which the agents take actions. Defaults to "simultaneous".
            model_name (str, optional): The name of the language model to use. Defaults to "gpt-3.5-turbo".
        """
        super().__init__()
        self.model_name = model_name
        if background_class is None:
            self.background_class = ScriptBackground
        else:
            self.background_class = background_class
        self.background = self.background_class(
            scenario="",
            p1_background="",
            p2_background="",
            p1_goal="",
            p2_goal="",
            p1_name="",
            p2_name="",
        )

        self.agents = []
        self.action_spaces = {}
        self.available_action_types = list(available_action_types)
        self.action_order = action_order
        self.action_mask: list[bool] = []
        self.evaluators = evaluators
        self.terminal_evaluators = terminal_evaluators

        # if an environment profile is provided, use it
        assert (
            env_profile or uuid_str
        ), "Either env_profile or uuid_str must be provided"
        if env_profile is not None:
            self.profile = env_profile
        # if a uuid is provided, try to load the environment profile from the database
        elif uuid_str is not None:
            # try retrieving profile from database
            try:
                self.profile = EnvironmentProfile.get(pk=uuid_str)
            except NotFoundError:
                raise ValueError(f"Agent with uuid {uuid_str} not found in database")

    @configurable
    def reset(
        self,
        seed: int | None = None,
        options: dict[str, str] | None = None,
        agents: Agents | None = None,
        omniscient: bool = False,
        lite: bool = False,
    ) -> dict[str, Observation]:
        """Starting a new episode. Must be called before step().

        Args:
            seed (int, optional): Seed for the environment. Defaults to None. Not used right now.
            options (dict, optional): Options for the environment. Defaults to None.
                "partial_background_file" (str): Path to a json file which need to contain a ScriptBackground object. The backgound can be incompleted ("unknown" for missing parts), and the missing parts will be filled in by the environment.
                "full_background_file" (str): Path to a json file which need to contain a ScriptBackground object. The backgound must be completed (no "unknown" for missing parts).
            omniscient (bool, optional): Whether the agents know the other agent's goal. Defaults to False.
        """
        super().__init__()
        MessengerMixin.reset_inbox(self)
        assert (
            not options
            or "partial_background_file" not in options
            and "full_background_file" not in options
        ), "partial_background_file and full_background_file are not supported anymore"
        if agents is not None:
            assert agents, "agents must be provided"
            assert len(agents) == 2, "Only supporting two agents right now"
            agent_names = list(agents.keys())
            agent_goals = self.profile.agent_goals
            assert len(agent_goals) == 2, "Only supporting two agents right now"

            raw_background = self.background_class(
                scenario=self.profile.scenario,
                p1_background=get_bio(
                    self.profile.relationship,
                    agents[agent_names[0]].profile,
                    agent_id=0,
                ),
                p2_background=get_bio(
                    self.profile.relationship,
                    agents[agent_names[1]].profile,
                    agent_id=1,
                ),
                p1_goal=f"<root viewer='agent_0'>{agent_goals[0]}</root>",
                p2_goal=f"<root viewer='agent_1'>{agent_goals[1]}</root>",
                p1_name=agent_names[0],
                p2_name=agent_names[1],
            )

            if lite:
                raw_background.p1_background = ""
                raw_background.p2_background = ""

            self.background = self.background_class(
                scenario=render_text_for_environment(raw_background.scenario),
                p1_background=render_text_for_environment(raw_background.p1_background),
                p2_background=render_text_for_environment(raw_background.p2_background),
                p1_goal=render_text_for_environment(raw_background.p1_goal),
                p2_goal=render_text_for_environment(raw_background.p2_goal),
                p1_name=raw_background.p1_name,
                p2_name=raw_background.p2_name,
            )
        else:
            raise ValueError("agents must be provided")

        self.agents = [self.background.p1_name, self.background.p2_name]
        agent_backgrounds = []
        if omniscient:
            for i in range(self.num_agents):
                agent_backgrounds.append(copy.deepcopy(self.background))
        else:
            for i in range(self.num_agents):
                agent_backgrounds.append(
                    self.background_class(
                        scenario=render_text_for_agent(raw_background.scenario, i),
                        p1_background=render_text_for_agent(
                            raw_background.p1_background, i
                        ),
                        p2_background=render_text_for_agent(
                            raw_background.p2_background, i
                        ),
                        p1_goal=render_text_for_agent(raw_background.p1_goal, i),
                        p2_goal=render_text_for_agent(raw_background.p2_goal, i),
                        p1_name=raw_background.p1_name,
                        p2_name=raw_background.p2_name,
                    )
                )
        background_for_a = agent_backgrounds[0]
        background_for_b = agent_backgrounds[1]

        if not omniscient:
            background_for_a.p2_goal = "Unknown"
            background_for_b.p1_goal = "Unknown"

        self.action_spaces = {
            agent: Dict(
                dict(
                    action_type=Discrete(len(self.available_action_types)),
                    argument=Text(256),
                )
            )
            for agent in self.agents
        }
        self.turn_number = 0
        self.action_mask = [False for _ in self.agents]
        if self.action_order == "round-robin":
            self.action_mask[0] = True
        elif self.action_order == "random":
            self.action_mask[random.randint(0, len(self.action_mask) - 1)] = True
        else:
            self.action_mask = [True for _ in self.agents]

        self.recv_message("Environment", self.background)

        return {
            self.background.p1_name: Observation(
                last_turn=background_for_a.to_natural_language(),
                turn_number=0,
                available_actions=list(self.available_action_types)
                if self.action_mask[0]
                else ["none"],
            ),
            self.background.p2_name: Observation(
                last_turn=background_for_b.to_natural_language(),
                turn_number=0,
                available_actions=list(self.available_action_types)
                if self.action_mask[1]
                else ["none"],
            ),
        }

    @beartype
    def step(
        self, actions: dict[str, AgentAction] | dict[str, dict[str, int | str]]
    ) -> tuple[
        dict[str, Observation],
        dict[str, float],
        dict[str, bool],
        dict[str, bool],
        dict[str, dict[Any, Any]],
    ]:
        # Time step ++
        self.turn_number += 1

        # For action sampled from action space, it needs to be converted into AgentAction
        complied_actions: dict[str, AgentAction] = {}
        for key in actions.keys():
            action = actions[key]
            if isinstance(action, AgentAction):
                complied_actions[key] = action
            else:
                action["action_type"] = self.available_action_types[
                    int(action["action_type"])
                ]
                complied_actions[key] = AgentAction.parse_obj(action)

        # Masking actions from agent that are in turn
        for idx, agent in enumerate(self.agents):
            if not self.action_mask[idx]:
                complied_actions[agent] = AgentAction(action_type="none", argument="")

        self.recv_message(
            "Environment", SimpleMessage(message=f"Turn #{self.turn_number}")
        )
        for agent, action in complied_actions.items():
            self.recv_message(agent, action)

        response = unweighted_aggregate_evaluate(
            list(
                itertools.chain(
                    *(
                        evaluator(turn_number=self.turn_number, messages=self.inbox)
                        for evaluator in self.evaluators
                    )
                )
            )
        )

        self.action_mask = [False for _ in self.agents]
        if self.action_order == "round-robin":
            self.action_mask[self.turn_number % len(self.action_mask)] = True
        elif self.action_order == "random":
            self.action_mask[random.randint(0, len(self.action_mask) - 1)] = True
        else:
            self.action_mask = [True for _ in self.agents]
        obs = _actions_to_natural_language(complied_actions)
        return (
            {
                self.background.p1_name: Observation(
                    last_turn=render_text_for_agent(obs, agent_id=0),
                    turn_number=self.turn_number,
                    available_actions=list(self.available_action_types)
                    if self.action_mask[0]
                    else ["none"],
                ),
                self.background.p2_name: Observation(
                    last_turn=render_text_for_agent(obs, agent_id=1),
                    turn_number=self.turn_number,
                    available_actions=list(self.available_action_types)
                    if self.action_mask[1]
                    else ["none"],
                ),
            },
            {
                self.background.p1_name: (
                    response.p1_rate
                    if isinstance(response.p1_rate, float)
                    else response.p1_rate[0]
                )
                if response.p1_rate
                else 0,
                self.background.p2_name: (
                    response.p2_rate
                    if isinstance(response.p2_rate, float)
                    else response.p2_rate[0]
                )
                if response.p2_rate
                else 0,
            },
            {
                self.background.p1_name: response.terminated,
                self.background.p2_name: response.terminated,
            },
            {
                self.background.p1_name: False,
                self.background.p2_name: False,
            },
            {
                self.background.p1_name: {
                    "comments": response.comments or "",
                    "complete_rating": response.p1_rate or 0,
                },
                self.background.p2_name: {
                    "comments": response.comments or "",
                    "complete_rating": response.p2_rate or 0,
                },
            },
        )

    @beartype
    async def astep(
        self, actions: dict[str, AgentAction] | dict[str, dict[str, int | str]]
    ) -> tuple[
        dict[str, Observation],
        dict[str, float],
        dict[str, bool],
        dict[str, bool],
        dict[str, dict[Any, Any]],
    ]:
        # Time step ++
        self.turn_number += 1

        # For action sampled from action space, it needs to be converted into AgentAction
        complied_actions: dict[str, AgentAction] = {}
        for key in actions.keys():
            action = actions[key]
            if isinstance(action, AgentAction):
                complied_actions[key] = action
            else:
                action["action_type"] = self.available_action_types[
                    int(action["action_type"])
                ]
                complied_actions[key] = AgentAction.parse_obj(action)

        # Masking actions from agent that are in turn
        for idx, agent in enumerate(self.agents):
            if not self.action_mask[idx]:
                complied_actions[agent] = AgentAction(action_type="none", argument="")

        self.recv_message(
            "Environment", SimpleMessage(message=f"Turn #{self.turn_number}")
        )
        for agent, action in complied_actions.items():
            self.recv_message(agent, action)

        response = unweighted_aggregate_evaluate(
            list(
                itertools.chain(
                    *await asyncio.gather(
                        *[
                            evaluator.__acall__(
                                turn_number=self.turn_number,
                                messages=self.inbox,
                            )
                            for evaluator in self.evaluators
                        ]
                    )
                )
            )
        )

        if response.terminated:
            terminal_response = unweighted_aggregate_evaluate(
                list(
                    itertools.chain(
                        *await asyncio.gather(
                            *[
                                evaluator.__acall__(
                                    turn_number=self.turn_number,
                                    messages=self.inbox,
                                )
                                for evaluator in self.terminal_evaluators
                            ]
                        )
                    )
                )
            )
            # incorporate terminal response into response
            response.p1_rate = response.p1_rate or terminal_response.p1_rate
            response.p2_rate = response.p2_rate or terminal_response.p2_rate
            if response.comments and terminal_response.comments:
                response.comments += terminal_response.comments
            elif terminal_response.comments:
                response.comments = terminal_response.comments

        self.action_mask = [False for _ in self.agents]
        if self.action_order == "round-robin":
            self.action_mask[self.turn_number % len(self.action_mask)] = True
        elif self.action_order == "random":
            self.action_mask[random.randint(0, len(self.action_mask) - 1)] = True
        else:
            self.action_mask = [True for _ in self.agents]
        obs = _actions_to_natural_language(complied_actions)
        info = {
            self.background.p1_name: {
                "comments": response.comments or "",
                "complete_rating": response.p1_rate or 0,
            },
            self.background.p2_name: {
                "comments": response.comments or "",
                "complete_rating": response.p2_rate or 0,
            },
        }
        if response.terminated:
            info["rewards_prompt"] = {
                "overall_prompt": self.terminal_evaluators[0].prompt  # type: ignore
            }

        return (
            {
                self.background.p1_name: Observation(
                    last_turn=render_text_for_agent(obs, agent_id=0),
                    turn_number=self.turn_number,
                    available_actions=list(self.available_action_types)
                    if self.action_mask[0]
                    else ["none"],
                ),
                self.background.p2_name: Observation(
                    last_turn=render_text_for_agent(obs, agent_id=1),
                    turn_number=self.turn_number,
                    available_actions=list(self.available_action_types)
                    if self.action_mask[1]
                    else ["none"],
                ),
            },
            {
                self.background.p1_name: (
                    response.p1_rate
                    if isinstance(response.p1_rate, float)
                    else response.p1_rate[0]
                )
                if response.p1_rate
                else 0,
                self.background.p2_name: (
                    response.p2_rate
                    if isinstance(response.p2_rate, float)
                    else response.p2_rate[0]
                )
                if response.p2_rate
                else 0,
            },
            {
                self.background.p1_name: response.terminated,
                self.background.p2_name: response.terminated,
            },
            {
                self.background.p1_name: False,
                self.background.p2_name: False,
            },
            info,
        )

    def render(self, mode: str = "human") -> None:
        pass

    def close(self) -> None:
        pass



from .message_classes import (
    ActionType,
    AgentAction,
    Message,
    Observation,
    ScriptBackground,
    ScriptEnvironmentResponse,
    SimpleMessage,
)
from .messenger import MessengerMixin

__all__ = [
    "Message",
    "Observation",
    "ScriptBackground",
    "ScriptEnvironmentResponse",
    "AgentAction",
    "ActionType",
    "SimpleMessage",
    "MessengerMixin",
]



# Intially created by:
# gbPagano <guilhermebpagano@gmail.com>
# https://github.com/gbPagano/rich_menu

from typing import Literal
import click
from rich.align import Align
from rich.console import Group
from rich.live import Live
from rich.panel import Panel
from rich.rule import Rule
from rich.text import Text


class Menu:
    def __init__(
        self,
        *options: str,
        start_index: int = 0,
        rule_title: str = "MENU",
        panel_title: str = "",
        color: str = "bold green",
        align: Literal["left", "center", "right"] = "center",
        selection_char: str = ">",
    ):
        self.options = options
        self.index = start_index
        self.rule_tile = rule_title
        self.panel_title = panel_title
        self.color = color
        self.align = align
        self.selection_char = selection_char

    def _get_click(self) -> str | None:
        match click.getchar():
            case "\r":
                return "enter"
            case "\x1b[B" | "s" | "S" | "àP":
                return "down"
            case "\x1b[A" | "w" | "W" | "àH":
                return "up"
            case "\x1b[D" | "a" | "A" | "àK":
                return "left"
            case "\x1b[C" | "d" | "D" | "àM":
                return "right"
            case "\x1b":
                return "exit"
            case _:
                return None

    def _update_index(self, key: str) -> None:
        if key == "down":
            self.index += 1
        elif key == "up":
            self.index -= 1

        if self.index > len(self.options) - 1:
            self.index = 0
        elif self.index < 0:
            self.index = len(self.options) - 1

    @property
    def _group(self) -> Group:
        menu = Text(justify="left")

        selected = Text(self.selection_char + " ", self.color)
        not_selected = Text(" " * (len(self.selection_char) + 1))
        selections = [not_selected] * len(self.options)
        selections[self.index] = selected

        for idx, option in enumerate(self.options):
            menu.append(Text.assemble(selections[idx], option + "\n"))
        menu.rstrip()

        panel = Panel.fit(menu)
        panel.title = Text(self.panel_title, self.color)
        if self.rule_tile:
            group = Group(
                Rule(self.rule_tile, style=self.color),
                Align(panel, self.align),
            )
        else:
            group = Group(
                Align(panel, self.align),
            )

        return group

    def _clean_menu(self) -> None:
        rule = 1 if self.rule_tile else 0
        for _ in range(len(self.options) + rule + 2):
            print("\x1b[A\x1b[K", end="")

    def ask(
        self, screen: bool = True, esc: bool = True, return_index: bool = False
    ) -> str | int:
        with Live(self._group, auto_refresh=False, screen=screen) as live:
            live.update(self._group, refresh=True)
            while True:
                try:
                    key = self._get_click()
                    if key == "enter":
                        break
                    elif key == "exit" and esc:
                        exit()
                    elif key in ["up", "down"]:
                        self._update_index(key)
                        live.update(self._group, refresh=True)
                except (KeyboardInterrupt, EOFError):
                    exit()

        if not screen:
            self._clean_menu()

        return self.options[self.index] if not return_index else self.index



from typing import Any

from pydantic import root_validator
from redis_om import JsonModel
from redis_om.model.model import Field

from sotopia.database.persistent_profile import AgentProfile


class EpisodeLog(JsonModel):
    # Note that we did not validate the following constraints:
    # 1. The number of turns in messages and rewards should be the same or off by 1
    # 2. The agents in the messages are the same as the agetns

    environment: str = Field(index=True)
    agents: list[str] = Field(index=True)
    tag: str | None = Field(index=True)
    models: list[str] | None = Field(index=True)
    messages: list[list[tuple[str, str, str]]]  # Messages arranged by turn
    reasoning: str
    rewards: list[tuple[float, dict[str, float]] | float]  # Rewards arranged by turn
    rewards_prompt: str

    @root_validator(skip_on_failure=True)
    def agent_number_message_number_reward_number_turn_number_match(
        cls, values: Any
    ) -> Any:
        agents, _, _reasoning, rewards = (
            values.get("agents"),
            values.get("messages"),
            values.get("reasoning"),
            values.get("rewards"),
        )
        agent_number = len(agents)

        assert (
            len(rewards) == agent_number
        ), f"Number of agents in rewards {len(rewards)} and agents {agent_number} do not match"
        return values

    def render_for_humans(self) -> tuple[list[AgentProfile], list[str]]:
        """Generate a human readable version of the episode log.

        Returns:
            A tuple of (a list of agent_profiles, a list of str): The agent profiles, and the messages and rewards in each turn.
        """

        agent_profiles = [AgentProfile.get(pk=uuid_str) for uuid_str in self.agents]
        messages_and_rewards = []
        for idx, turn in enumerate(self.messages):
            messages_in_this_turn = []
            if idx == 0:
                assert (
                    len(turn) >= 2
                ), "The first turn should have at least environemnt messages"
                messages_in_this_turn.append(turn[0][2])
                messages_in_this_turn.append(turn[1][2])
            for sender, receiver, message in turn:
                if receiver == "Environment":
                    if sender != "Environment":
                        if "did nothing" in message:
                            continue
                        else:
                            if "said:" in message:
                                messages_in_this_turn.append(f"{sender} {message}")
                            else:
                                messages_in_this_turn.append(f"{sender}: {message}")
                    else:
                        messages_in_this_turn.append(message)
            messages_and_rewards.append("\n".join(messages_in_this_turn))
        messages_and_rewards.append(f"The reasoning is:\n{self.reasoning}")
        messages_and_rewards.append(
            f"The rewards are:\nAgent 1: {self.rewards[0]}\nAgent 2: {self.rewards[1]}"
        )
        return agent_profiles, messages_and_rewards


class AnnotationForEpisode(JsonModel):
    episode: str = Field(index=True, description="the pk id of episode log")
    annotator_id: str = Field(index=True, full_text_search=True)
    rewards: list[tuple[float, dict[str, float]] | float]
    reasoning: str



import asyncio
import itertools
import logging
from typing import Literal, Sequence, Type, cast

import gin
import rich
from beartype import beartype

from sotopia.agents import (
    Agents,
    HumanAgent,
    LLMAgent,
    RedisAgent,
    ScriptWritingAgent,
)
from sotopia.agents.base_agent import BaseAgent
from sotopia.database import EpisodeLog
from sotopia.envs import ParallelSotopiaEnv
from sotopia.envs.evaluators import (
    EvaluationForTwoAgents,
    ReachGoalLLMEvaluator,
    RuleBasedTerminatedEvaluator,
    SotopiaDimensions,
    unweighted_aggregate_evaluate,
)
from sotopia.generation_utils.generate import LLM_Name, agenerate_script
from sotopia.messages import AgentAction, Message, Observation
from sotopia.messages.message_classes import (
    ScriptBackground,
    ScriptEnvironmentResponse,
)
from sotopia.samplers import BaseSampler, EnvAgentCombo


@beartype
def run_sync_server(
    model_name_dict: dict[str, LLM_Name],
    action_order: Literal["simultaneous", "round-robin", "random"],
    agents_info: dict[str, dict[str, str]] | None = None,
    partial_background_file: str | None = None,
    full_background_file: str | None = None,
    mode: str | None = None,
) -> list[tuple[str, str, Message]]:
    # Create Environment and agents
    # This step will be moved to outside this function

    env = ParallelSotopiaEnv(
        model_name=model_name_dict["env"],
        action_order=action_order,
        evaluators=[
            RuleBasedTerminatedEvaluator(),
        ],
    )
    if partial_background_file:
        environment_messages = env.reset(
            options={"partial_background_file": partial_background_file}
        )
    elif full_background_file:
        environment_messages = env.reset(
            options={"full_background_file": full_background_file}
        )
    else:
        environment_messages = env.reset()
    agents = Agents()
    agents_model_names = [model_name_dict["agent1"], model_name_dict["agent2"]]
    for agent_name, agent_model in zip(env.agents, agents_model_names):
        if agent_model == "human":
            agents[agent_name] = HumanAgent(agent_name)
        elif mode == "speak":
            raise NotImplementedError(
                "Deprecated. The original Speaker Agent is not implemented in the async context."
            )
        else:
            agents[agent_name] = LLMAgent(agent_name, model_name=agent_model)
    agents.reset()

    messages: list[tuple[str, str, Message]] = []

    # Main Event Loop
    done = False
    for agent_name in env.agents:
        messages.append(("Environment", agent_name, environment_messages[agent_name]))

    while not done:
        # gather agent messages
        agent_messages: dict[str, AgentAction] = dict()
        for agent_name in env.agents:
            if agents_info is not None:
                agents[agent_name].goal = agents_info[agent_name]["goal"]
            agent_messages[agent_name] = agents[agent_name].act(
                environment_messages[agent_name]
            )
            messages.append((agent_name, "Environment", agent_messages[agent_name]))

        # send agent messages to environment
        environment_messages, _, terminated, ___, ____ = env.step(agent_messages)
        for agent_name in env.agents:
            messages.append(
                ("Environment", agent_name, environment_messages[agent_name])
            )
        done = all(terminated.values())

    return messages


@gin.configurable
async def arun_one_episode(
    env: ParallelSotopiaEnv,
    agent_list: Sequence[BaseAgent[Observation, AgentAction]],
    omniscient: bool = False,
    script_like: bool = False,
    json_in_script: bool = False,
    tag: str | None = None,
    push_to_db: bool = False,
) -> list[tuple[str, str, Message]]:
    agents = Agents({agent.agent_name: agent for agent in agent_list})
    environment_messages = env.reset(agents=agents, omniscient=omniscient)
    agents.reset()

    messages: list[list[tuple[str, str, Message]]] = []

    # Main Event Loop
    done = False
    messages.append(
        [
            ("Environment", agent_name, environment_messages[agent_name])
            for agent_name in env.agents
        ]
    )
    # set goal for agents
    for index, agent_name in enumerate(env.agents):
        agents[agent_name].goal = env.profile.agent_goals[index]
    rewards: list[list[float]] = []
    reasons: list[str] = []
    while not done:
        # gather agent messages
        agent_messages: dict[str, AgentAction] = dict()
        actions = await asyncio.gather(
            *[
                agents[agent_name].aact(environment_messages[agent_name])
                for agent_name in env.agents
            ]
        )
        if script_like:
            # manually mask one message
            agent_mask = env.action_mask
            for idx in range(len(agent_mask)):
                print("Current mask: ", agent_mask)
                if agent_mask[idx] == 0:
                    print("Action not taken: ", actions[idx])
                    actions[idx] = AgentAction(action_type="none", argument="")
                else:
                    print("Current action taken: ", actions[idx])

        # actions = cast(list[AgentAction], actions)
        for idx, agent_name in enumerate(env.agents):
            agent_messages[agent_name] = actions[idx]

            messages[-1].append((agent_name, "Environment", agent_messages[agent_name]))

        # send agent messages to environment
        (
            environment_messages,
            rewards_in_turn,
            terminated,
            ___,
            info,
        ) = await env.astep(agent_messages)
        messages.append(
            [
                ("Environment", agent_name, environment_messages[agent_name])
                for agent_name in env.agents
            ]
        )
        # print("Environment message: ", environment_messages)
        # exit(0)
        rewards.append([rewards_in_turn[agent_name] for agent_name in env.agents])
        reasons.append(
            " ".join(info[agent_name]["comments"] for agent_name in env.agents)
        )
        done = all(terminated.values())

    # TODO: clean up this part
    epilog = EpisodeLog(
        environment=env.profile.pk,
        agents=[agent.profile.pk for agent in agent_list],
        tag=tag,
        models=[env.model_name, agent_list[0].model_name, agent_list[1].model_name],
        messages=[
            [(m[0], m[1], m[2].to_natural_language()) for m in messages_in_turn]
            for messages_in_turn in messages
        ],
        reasoning=info[env.agents[0]]["comments"],
        rewards=[info[agent_name]["complete_rating"] for agent_name in env.agents],
        rewards_prompt=info["rewards_prompt"]["overall_prompt"],
    )
    rich.print(epilog.rewards_prompt)
    agent_profiles, conversation = epilog.render_for_humans()
    for agent_profile in agent_profiles:
        rich.print(agent_profile)
    for message in conversation:
        rich.print(message)

    if push_to_db:
        try:
            epilog.save()
        except Exception as e:
            logging.error(f"Failed to save episode log: {e}")
    # flatten nested list messages
    return list(itertools.chain(*messages))


@gin.configurable
@beartype
async def run_async_server(
    sampler: BaseSampler[Observation, AgentAction] = BaseSampler(),
    action_order: Literal["simutaneous", "round-robin", "random"] = "round-robin",
    model_dict: dict[str, LLM_Name] = {},
    env_agent_combo_list: list[EnvAgentCombo[Observation, AgentAction]] = [],
    omniscient: bool = False,
    script_like: bool = False,
    json_in_script: bool = False,
    tag: str | None = None,
    push_to_db: bool = False,
    using_async: bool = True,
) -> list[list[tuple[str, str, Message]]]:
    """
    Doc incomplete

    Args:
        omniscient (bool): Whether the agent knows the goal of the other, default to False
        script_like (bool): Whether we generate the turn in script like manner, default to False
        json_in_script (bool): Whether we requires the script generator to return json (Only valid when script_like is True), default to False

    Note: env_agent_combo_list is optional. When it defaults to [], sampler is used
    else the sampler is not used. Please pass in BaseSampler or simply not specify it when using this option.
    """

    assert not (push_to_db and tag is None), "please provide a tag when push to db"
    assert (
        model_dict or env_agent_combo_list
    ), "please provide model_dict or env_agent_combo_list"

    # Create Environment and agents
    # This step will be moved to outside this function

    def get_agent_class(
        model_name: str,
    ) -> Type[BaseAgent[Observation, AgentAction]]:
        if model_name == "human":
            return HumanAgent
        elif model_name == "redis":
            return RedisAgent
        elif script_like and not json_in_script:
            return ScriptWritingAgent
        else:
            return LLMAgent

    if env_agent_combo_list:
        assert (
            type(sampler) is BaseSampler
        ), "No sampler should be used when `env_agent_combo_list` is not empty"
        env_agent_combo_iter = iter(env_agent_combo_list)
    else:
        env_params = {
            "model_name": model_dict["env"],
            "action_order": action_order,
            "evaluators": [
                RuleBasedTerminatedEvaluator(max_turn_number=20, max_stale_turn=2),
            ],
            "terminal_evaluators": [
                ReachGoalLLMEvaluator(
                    model_dict["env"],
                    EvaluationForTwoAgents[SotopiaDimensions],
                ),
            ],
        }
        agents_model_dict = {
            "agent1": model_dict["agent1"],
            "agent2": model_dict["agent2"],
        }
        env_agent_combo_iter = sampler.sample(
            agent_classes=[
                get_agent_class(model_name) for model_name in agents_model_dict.values()
            ],
            n_agent=len(agents_model_dict),
            env_params=env_params,
            agents_params=[
                {"model_name": model_name} if model_name != "human" else {}
                for model_name in agents_model_dict.values()
            ],
        )
    episode_futures = [
        arun_one_episode(
            env=env_agent_combo[0],
            agent_list=env_agent_combo[1],
            omniscient=omniscient,
            script_like=script_like,
            json_in_script=json_in_script,
            tag=tag,
            push_to_db=push_to_db,
        )
        for env_agent_combo in env_agent_combo_iter
    ]

    batch_results = (
        await asyncio.gather(*episode_futures)
        if using_async
        else [await i for i in episode_futures]
    )

    return cast(list[list[tuple[str, str, Message]]], batch_results)


async def arun_one_script(
    env: ParallelSotopiaEnv,
    agent_list: Sequence[BaseAgent[Observation, AgentAction]],
    model_dict: dict[str, LLM_Name],
    omniscient: bool = False,
    tag: str | None = None,
    push_to_db: bool = False,
) -> list[tuple[str, str, Message]]:
    """
    Generate script for one episode
    Args:
        omniscient (bool): Whether the agent knows the goal of the other
    """

    agents = Agents({agent.agent_name: agent for agent in agent_list})
    env.reset(agents=agents, omniscient=omniscient)

    agent_names = [agent.agent_name for agent in agent_list]
    assert len(agent_names) == 2, f"only support 2 agents, current: {agent_names}"

    script_background = env.inbox[0][1]
    assert isinstance(script_background, ScriptBackground)
    story, prompt = await agenerate_script(
        model_name=model_dict["env"],
        background=script_background,
        agent_names=agent_names,
    )
    messages, agent_messages = story
    env_message = [("Environment", script_background)]
    agent_messages = env_message + agent_messages

    evaluator = ReachGoalLLMEvaluator(
        model_name="gpt-4",
        response_format_class=EvaluationForTwoAgents[SotopiaDimensions],
    )
    response = unweighted_aggregate_evaluate(
        list(
            itertools.chain(
                *await asyncio.gather(
                    *[
                        sing_evaluator.__acall__(
                            turn_number=-1,
                            messages=agent_messages,
                        )
                        for sing_evaluator in [evaluator]
                    ]
                )
            )
        )
    )
    info: dict[str, dict[str, str | ScriptEnvironmentResponse | float | None]] = {
        script_background.p1_name: {
            "comments": response.comments or "",
            "complete_rating": response.p1_rate or 0,  # type: ignore
        },
        script_background.p2_name: {
            "comments": response.comments or "",
            "complete_rating": response.p2_rate or 0,  # type: ignore
        },
        "rewards_prompt": {"overall_prompt": evaluator.prompt or ""},
    }
    epilog = EpisodeLog(
        environment=env.profile.pk,
        agents=[agent.profile.pk for agent in agent_list],
        tag=tag,
        models=[model_dict["env"], model_dict["agent1"], model_dict["agent2"]],
        messages=[
            [(m[0], m[1], m[2].to_natural_language()) for m in messages_in_turn]
            for messages_in_turn in messages
        ],
        reasoning=str(info[env.agents[0]]["comments"])
        + str(info[env.agents[1]]["comments"]),
        rewards=[info[agent_name]["complete_rating"] for agent_name in env.agents],
        rewards_prompt=info["rewards_prompt"]["overall_prompt"],
    )
    print("Reward prompt: ")
    rich.print(epilog.rewards_prompt)
    agent_profiles, conversation = epilog.render_for_humans()
    print("Agent profiles: ")
    for agent_profile in agent_profiles:
        rich.print(agent_profile)
    for message in conversation:
        rich.print(message)

    if push_to_db:
        try:
            epilog.save()
        except Exception as e:
            logging.error(f"Failed to save episode log: {e}")
    # flatten nested list messages
    return list(itertools.chain(*messages))


async def aevaluate_one_episode(
    episode: EpisodeLog,
    model: str = "gpt-4",
    tag: str | None = None,
    push_to_db: bool = False,
) -> None:
    history = episode.rewards_prompt.replace("Prompt after formatting:", "").split(
        ",\nBased on previous interactions"
    )[0]
    evaluator = ReachGoalLLMEvaluator(
        model_name=model,
        response_format_class=EvaluationForTwoAgents[SotopiaDimensions],
    )
    response = unweighted_aggregate_evaluate(
        list(
            itertools.chain(
                *await asyncio.gather(
                    *[
                        single_evaluator.__acall__(
                            turn_number=-1,
                            history=history,
                            messages=None,
                            temperature=0.0,
                        )
                        for single_evaluator in [evaluator]
                    ]
                )
            )
        )
    )
    info: dict[str, dict[str, str | ScriptEnvironmentResponse | float | None]] = {
        episode.agents[0]: {
            "comments": response.comments or "",
            "complete_rating": response.p1_rate or 0,  # type: ignore
        },
        episode.agents[1]: {
            "comments": response.comments or "",
            "complete_rating": response.p2_rate or 0,  # type: ignore
        },
    }
    assert isinstance(episode.models, list)
    epilog = EpisodeLog(
        environment=episode.environment,
        agents=episode.agents,
        tag=tag,
        models=[model, episode.models[1], episode.models[2]],
        messages=episode.messages,
        reasoning=str(info[episode.agents[0]]["comments"])
        + str(info[episode.agents[1]]["comments"]),
        rewards=[info[agent_name]["complete_rating"] for agent_name in episode.agents],
        rewards_prompt="TBD",
    )
    # rich.print(history)
    # rich.print(epilog.rewards)

    if push_to_db:
        try:
            epilog.save()
        except Exception as e:
            logging.error(f"Failed to save episode log: {e}")



import re
from typing import Literal, cast

from pydantic import BaseModel, Field

from sotopia.utils import format_docstring

ActionType = Literal["none", "speak", "non-verbal communication", "action", "leave"]


class Message(BaseModel):
    """
    An interface for messages.
    There is only one required method: to_natural_language
    """

    def to_natural_language(self) -> str:
        raise NotImplementedError


class SimpleMessage(Message):
    """
    A simple message with a single string field.
    """

    message: str = Field(description="the message")

    def to_natural_language(self) -> str:
        return self.message


class Observation(Message):
    last_turn: str = Field(description="the last turn of the conversation")
    turn_number: int = Field(description="the turn number of the conversation")
    available_actions: list[ActionType] = Field(description="the available actions")

    def to_natural_language(self) -> str:
        if self.turn_number == 0:
            return f"\n{self.last_turn}\nConversation Starts:\n"
        else:
            return f"Turn #{self.turn_number-1}: {self.last_turn}\n"


class ScriptBackground(Message):
    scenario: str = Field(description="scenario of the episode")
    p1_name: str = Field(description="name of participant 1")
    p2_name: str = Field(description="name of participant 2")
    p1_background: str = Field(description="background of participant 1")
    p2_background: str = Field(description="background of participant 2")
    p1_goal: str = Field(description="goal of participant 1")
    p2_goal: str = Field(description="goal of participant 2")

    def to_natural_language(self) -> str:
        if self.p1_background or self.p2_background:
            p1_background = self.p1_background if self.p1_background else "Unknown"
            p2_background = self.p2_background if self.p2_background else "Unknown"
            # Not using AND, since in stranger relation the background is not visible
            return format_docstring(
                f"""Here is the context of this interaction:
            Scenario: {self.scenario}
            Participants: {self.p1_name} and {self.p2_name}
            {self.p1_name}'s background: {p1_background}
            {self.p2_name}'s background: {p2_background}
            {self.p1_name}'s goal: {self.p1_goal}
            {self.p2_name}'s goal: {self.p2_goal}
            """
            )
        else:
            return format_docstring(
                f"""Here is the context of this interaction:
            Scenario: {self.scenario}
            Participants: {self.p1_name} and {self.p2_name}
            {self.p1_name}'s goal: {self.p1_goal}
            {self.p2_name}'s goal: {self.p2_goal}
            """
            )


class ScriptEnvironmentResponse(Message):
    terminated: bool = Field(
        description="whether the conversation is terminated",
        default_factory=lambda: False,
    )
    p1_rate: float | tuple[float, dict[str, float]] | None = Field(
        description="rating of participant 1, on the scale of 1 to 10"
    )
    p2_rate: float | tuple[float, dict[str, float]] | None = Field(
        description="rating of participant 2, on the scale of 1 to 10"
    )
    comments: str | None = Field(
        description="All of the comments supporting the termination and rating"
    )

    def to_natural_language(self) -> str:
        reason_to_stop = format_docstring(
            f"""Environment response:
        {"The conversation is terminated." if self.terminated else ""}
        {"Rating of participant 1" + str(self.p1_rate) if self.p1_rate is not None else ""}
        {"Rating of participant 2" + str(self.p2_rate) if self.p2_rate is not None else ""}
        {self.comments if self.comments is not None else ""}
        """
        )
        clean_text = ""
        for line in reason_to_stop.split("\n"):
            if line.strip():
                clean_text += line + "\n"
        return clean_text


class AgentAction(Message):
    action_type: ActionType = Field(
        description="whether to speak at this turn or choose to not do anything"
    )
    argument: str = Field(
        description="the utterance if choose to speak, the expression or gesture if choose non-verbal communication, or the physical action if choose action"
    )

    def to_natural_language(self) -> str:
        match self.action_type:
            case "none":
                return "did nothing"
            case "speak":
                return f'said: "{self.argument}"'
            case "non-verbal communication":
                return f"[{self.action_type}] {self.argument}"
            case "action":
                return f"[{self.action_type}] {self.argument}"
            case "leave":
                return "left the conversation"


ScriptInteractionReturnType = tuple[
    list[list[tuple[str, str, Message]]], list[tuple[str, Message]]
]


class ScriptInteraction(Message):
    interactions: str = Field(
        description="""The interaction between the two participants in maximum 20 turns. Each turn is separated by a newline, and should only describe one agent. Following the structure:
        Turn #x
        [participant's name] [action] {argument for some actions}

        You can use different types of actions, but only use one in each turn. You should move other information into argument part. Below shows a python code snippet of the format for each action type:
        match self.action_type:
            case "none":
                return "did nothing"
            case "speak":
                return f'said: "{self.argument}"'
            case "non-verbal communication":
                return f"[{self.action_type}] {self.argument}"
            case "action":
                return f"[{self.action_type}] {self.argument}"
            case "leave":
                return "left the conversation"

        For example, the following is acceptable:
        Turn #x
        Oliver Thompson said: "Hey Esmeralda, what's wrong? You seem upset."
        Turn #x
        Esmeralda Solis [action] moved closer
        Turn #x
        Oliver Thompson [non-verbal communication] smiled
        Turn #x
        Esmeralda Solis did nothing
        Turn #x
        Oliver Thompson left the conversation
        Turn #x
        Esmeralda Solis [action] leaned in and lowered her voice: "Sorry"

        And the following is not acceptable:
        Turn #1
        Oliver Thompson [speak] said: "Hey Esmeralda, what's wrong? You seem upset."
        Turn #1
        Esmeralda Solis non-verbal communication moved closer
        """
    )

    def to_natural_language(self) -> str:
        return self.interactions

    def parse(
        self, agent_names: list[str], background: str
    ) -> tuple[list[list[tuple[str, str, Message]]], list[tuple[str, Message]]]:
        interaction = self.interactions
        # print("Interaction: ", interaction)
        lines = self.split_by_turn(interaction)

        agent_results = []
        results: list[list[tuple[str, str, Message]]] = [
            [
                (
                    "Environment",
                    name,
                    Observation(
                        last_turn=background,
                        turn_number=0,
                        available_actions=["none"],
                    ),
                )
                for name in agent_names
            ]
        ]

        for line_idx, line in enumerate(lines):
            try:
                res = self.parse_single_dialogue(line)
                action: AgentAction = cast(AgentAction, res["action"])
                argument: str = cast(str, res["argument"])
                cast(int, res["turn"])
                name: str = cast(str, res["name"])

                parsed_action = AgentAction(action_type=action, argument=argument)
                if name not in agent_names:
                    print(
                        f"The name of the agent, {name}, is not in the list of agent names, {agent_names}"
                    )
                    name = agent_names[
                        line_idx % 2
                    ]  # TODO Not sure what name to be set here
            except Exception as e:
                print(
                    f"Error when parsing the dialogue: {line}",
                    f"The error is: {e}",
                )
                raise e
                parsed_action = AgentAction(action_type="none", argument="")
                name = agent_names[line_idx % 2]  # TODO same question as above
            inactive_agent_name = (
                agent_names[0] if name == agent_names[1] else agent_names[1]
            )
            results.append(
                [
                    (
                        "Environment",
                        name,
                        Observation(
                            last_turn="environment is the agent",
                            turn_number=line_idx + 1,
                            available_actions=["none"],
                        ),
                    )
                    for name in agent_names
                ]
                + [
                    (name, "Environment", parsed_action),
                    (
                        inactive_agent_name,
                        "Environment",
                        AgentAction(action_type="none", argument="did nothing"),
                    ),
                ]
            )

            agent_results.append((name, parsed_action))
        # print("Parsed agent results: ", agent_results)
        return (results, agent_results)  # type: ignore

    def parse_single_dialogue(
        self, dialogue: str
    ) -> dict[str, str | int | AgentAction | None]:
        """Parse a single dialogue string and return a dictionary with turn, name, action, and argument."""

        # Match the turn number and name. Assume all agent name starts with a capital letter and is followed by lowercase letters
        match_turn_name = re.match(
            r"Turn #?(\d+):?\s*\n((?:[A-Z]['a-z]* ?)+)", dialogue
        )

        if not match_turn_name:
            raise ValueError(
                f"The dialogue does not match the expected format: {dialogue}"
            )
            return None  # TODO Which should we use, return None or raise error?

        turn, name = match_turn_name.groups()
        action_content = dialogue[
            len(match_turn_name.group(0)) :
        ].strip()  # Extract the action content

        # Check for different action types
        if "did nothing" in action_content:
            action, argument = "none", ""
        elif match := re.match(r'said: "(.*?)"', action_content):
            action, argument = "speak", match.group(1)
            action, argument = action.strip(), argument.strip()
        elif match := re.match(r'\[speak\] said: "(.*?)"', action_content):
            action, argument = "speak", match.group(1)
            action, argument = action.strip(), argument.strip()
        elif match := re.match(
            r"\[(non-verbal communication|action)\] (.*)", action_content
        ):
            action, argument = match.groups()
        elif "left the conversation" in action_content:
            # TODO Make it more elegant to handle the situation of `left the conversation.`
            action, argument = "leave", ""
        else:
            action, argument = None, None

        parsed_item = {
            "turn": int(turn),
            "name": name.strip(),
            "action": action,
            "argument": argument,
        }
        return parsed_item

    def split_by_turn(self, input_string: str) -> list[str]:
        """Split the input dialogue string by turn and return a list of dialogues."""
        # Split using 'Turn #' as delimiter, but keep the delimiter in the results
        dialogues = re.split(r"(?=Turn #?\d+)", input_string)
        # Remove any empty strings and strip whitespace
        dialogues = [dialogue.strip() for dialogue in dialogues if dialogue.strip()]
        dialogues = [dialogue for dialogue in dialogues if dialogue.startswith("Turn")]
        # Change from Turn #x to Turn (#)x (# is optional)
        dialogues[-1] = "\n".join(
            dialogues[-1].split("\n")[:2]
        )  # Discard further input in the last turn

        for dialogue in dialogues:
            # TODO this is current workaround for the issue of multiple agents in one turn
            if len(dialogue.split("\n")) >= 3:
                raise ValueError("Only one agent can act per turn.")
        return dialogues

    @staticmethod
    def default_value_for_return_type() -> ScriptInteractionReturnType:
        results_1: list[list[tuple[str, str, Message]]] = [
            [
                (
                    "Environment",
                    name,
                    Observation(
                        last_turn="Environment is the agent",
                        turn_number=0,
                        available_actions=["none"],
                    ),
                )
                for name in ["none", "none"]
            ]
        ]
        results_2: list[tuple[str, Message]] = [
            ("", AgentAction(action_type="none", argument=""))
        ]
        return (results_1, results_2)



from enum import IntEnum
from typing import Any

from pydantic import root_validator
from redis_om import JsonModel
from redis_om.model.model import Field


class RelationshipType(IntEnum):
    stranger = 0
    know_by_name = 1
    acquaintance = 2
    friend = 3
    romantic_relationship = 4
    family_member = 5


class AgentProfile(JsonModel):
    first_name: str = Field(index=True)
    last_name: str = Field(index=True)
    age: int = Field(index=True, default_factory=lambda: 0)
    occupation: str = Field(index=True, default_factory=lambda: "")
    gender: str = Field(index=True, default_factory=lambda: "")
    gender_pronoun: str = Field(index=True, default_factory=lambda: "")
    public_info: str = Field(index=True, default_factory=lambda: "")
    big_five: str = Field(index=True, default_factory=lambda: "")
    moral_values: list[str] = Field(index=False, default_factory=lambda: [])
    schwartz_personal_values: list[str] = Field(index=False, default_factory=lambda: [])
    personality_and_values: str = Field(index=True, default_factory=lambda: "")
    decision_making_style: str = Field(index=True, default_factory=lambda: "")
    secret: str = Field(default_factory=lambda: "")
    model_id: str = Field(default_factory=lambda: "")
    mbti: str = Field(default_factory=lambda: "")


class EnvironmentProfile(JsonModel):
    codename: str = Field(
        index=True,
        default_factory=lambda: "",
        description="The codename of the environment",
    )
    source: str = Field(
        index=True,
        default_factory=lambda: "",
        description="The source of the environment",
    )
    scenario: str = Field(
        index=True,
        default_factory=lambda: "",
        description="A concrete scenario of where the social interaction takes place, the scenario should have two agents (agent1 and agent2), and you should illustrate the relationship between the two agents, and for what purpose agent1 is interacting with agent2. Please avoid mentioning specific names and occupations in the scenario and keep all the mentions gender-neutral. Also avoid generating scenarios that requires childrend (below 18) or elderly (above 70) to be involved.",
    )
    agent_goals: list[str] = Field(
        default_factory=lambda: [],
        description="The social goals of each agent, which could include <extra_info>...</extra_info>, <clarification_hint>...</clarification_hint>, and <strategy_hint>...</strategy_hint> to help the agent achieve the goal. Avoid providing too specific strategy hint, try to be as abstract as possible. For example, use 'you can provide financial benefits to achieve your goal' instead of 'you can buy him a boba tea to achieve your goal.'",
    )
    relationship: RelationshipType = Field(
        index=True,
        default_factory=lambda: RelationshipType.stranger,
        description="The relationship between the two agents, choose from: stranger, know_by_name, acquaintance, friend, romantic_relationship, family_member. Do not make up a relationship, but choose from the list, 0 means stranger, 1 means know_by_name, 2 means acquaintance, 3 means friend, 4 means romantic_relationship, 5 means family_member",
    )
    age_constraint: str | None = Field(
        default_factory=lambda: None,
        description="The age constraint of the environment, a list of tuples, each tuple is a range of age, e.g., '[(18, 25), (30, 40)]' means the environment is only available to agent one between 18 and 25, and agent two between 30 and 40",
    )
    occupation_constraint: str | None = Field(
        default_factory=lambda: None,
        description="The occupation constraint of the environment, a list of lists, each list is a list of occupations, e.g., '[['student', 'teacher'], ['doctor', 'nurse']]' means the environment is only available to agent one if agent one is a student or a teacher, and agent two is a doctor or a nurse",
    )
    agent_constraint: list[list[str]] | None = Field(
        default_factory=lambda: None,
    )


class RelationshipProfile(JsonModel):
    agent_1_id: str = Field(index=True)
    agent_2_id: str = Field(index=True)
    relationship: RelationshipType = Field(
        index=True,
        description="0 means stranger, 1 means know_by_name, 2 means acquaintance, 3 means friend, 4 means romantic_relationship, 5 means family_member",
    )  # this could be improved by limiting str to a relationship Enum
    background_story: str | None = Field(default_factory=lambda: None)


class EnvironmentList(JsonModel):
    name: str = Field(index=True)
    environments: list[str] = Field(default_factory=lambda: [])
    agent_index: list[str] | None = Field(default_factory=lambda: None)

    # validate the length of agent_index should be same as environments
    @root_validator
    def the_length_agent_index_matches_environments(cls, values: Any) -> Any:
        environments, agent_index = (
            values.get("environments"),
            values.get("agent_index"),
        )
        if agent_index is None:
            return values
        assert (
            len(environments) == len(agent_index)
        ), f"Number of environments {len(environments)} and agent_index {len(agent_index)} do not match"
        return values
